/*************************************************************************************
*	irrlamb - http://code.google.com/p/irrlamb/
*	Copyright (C) 2007  Alan Witkowski
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation, either version 3 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program.  If not, see <http://www.gnu.org/licenses/>.
**************************************************************************************/
#include "camera.h"
#include "replay.h"
#include "physics.h"

// Constructor
CameraClass::CameraClass()
:	Node(NULL),
	Yaw(0.0f),
	Pitch(0.0f),
	MaxDistance(1.0f),
	Distance(0.0f) {

	Node = irrScene->addCameraSceneNode();
	Node->setNearValue(0.1f);
	Node->setFarValue(1000.0f);

}

// Destructor
CameraClass::~CameraClass() {

	Node->remove();
}

// Updates the camera based on the mouse
void CameraClass::HandleMouseMotion(float TUpdateX, float TUpdateY) {

	Yaw += TUpdateX;
	Pitch += TUpdateY;

	if(Pitch > 89.0f)
		Pitch = 89.0f;
	else if(Pitch <= -89.0f)
		Pitch = -89.0f;
}

// Updates the camera
void CameraClass::Update(const vector3df &TTarget) {
	
	// Get camera rotation
	Transform.makeIdentity();
    Transform.setRotationDegrees(vector3df(Pitch, Yaw, 0.0f)); 

	// Set distance from object
	Distance = MaxDistance;

	// Get camera offset
	vector3df Offset(0.0f, 0.0f, 1.0f);
	Transform.transformVect(Offset);

	// Save camera direction
	Direction = Offset;

	// Position the camera
	Offset *= -Distance;

	// Set camera target
	Node->setTarget(TTarget);

	// Get ray intervals
	btVector3 RayStart(TTarget.X, TTarget.Y, TTarget.Z), RayEnd(RayStart);
	RayEnd[0] += Offset.X;
	RayEnd[1] += Offset.Y;
	RayEnd[2] += Offset.Z;
	
	// Get point on wall where ray collides
	btVector3 HitNormal;
	Physics::Instance().RaycastWorld(RayStart, RayEnd, HitNormal);
	
	// Shorten the offset a little bit
	btVector3 NewOffset = RayEnd - RayStart;
	NewOffset *= 0.9f;

	//printf("%f %f %f\n", NewOffset[0], NewOffset[1], NewOffset[2]);
	//printf("%f %f %f\n", RayEnd[0], RayEnd[1], RayEnd[2]);

	// Set the new position
	RayEnd = RayStart + NewOffset;
	Node->setPosition(vector3df(RayEnd[0], RayEnd[1], RayEnd[2]));

	// Note changes
	if(!MovementChanged && (PreviousPosition != Node->getPosition() || PreviousLookAt != Node->getTarget()))
		MovementChanged = true;

	PreviousPosition = Node->getPosition();
	PreviousLookAt = Node->getTarget();
	
	// Write replay
	if(Replay::Instance().NeedsPacket() && MovementChanged) {
		MovementChanged = false;

		// Write the event
		Replay::Instance().WriteEvent(ReplayClass::PACKET_CAMERA);

		// Write packet
		ReplayCameraStruct Packet;
		Packet.Position = Node->getPosition();
		Packet.LookAt = Node->getTarget();
		Replay::Instance().WritePacket(Packet);
	}
}
