/*************************************************************************************
*	irrlamb - http://code.google.com/p/irrlamb/
*	Copyright (C) 2007  Alan Witkowski
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation, either version 3 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program.  If not, see <http://www.gnu.org/licenses/>.
**************************************************************************************/
#include "config.h"
#include "globals.h"
#include "save.h"
#include "../tinyxml/tinyxml.h"

// Initializes the config system
int ConfigClass::Init() {

	Reset();

	return 1;
}

// Closes the config system
int ConfigClass::Close() {
	
	return 1;
}

// Resets the configuration to the default values
void ConfigClass::Reset() {
	
	// Video
	DriverType = EDT_OPENGL;
	ScreenWidth = 800;
	ScreenHeight = 600;
	FullScreen = 0;

	// Audio
	AudioEnabled = 0;
	SoundVolume = 1.0;
	MusicVolume = 1.0;

	// Input
	Keys[ACTIONS::MOVEFORWARD] = KEY_KEY_E;
	Keys[ACTIONS::MOVEBACK] = KEY_KEY_D;
	Keys[ACTIONS::MOVELEFT] = KEY_KEY_S;
	Keys[ACTIONS::MOVERIGHT] = KEY_KEY_F;
	Keys[ACTIONS::RESET] = KEY_KEY_X;
	Keys[ACTIONS::JUMP] = KEY_SPACE;

	MouseScaleX = 1.0f;
	MouseScaleY = 1.0f;

	// Replays
	ReplayInterval = 15;
}

// Reads the config file
int ConfigClass::ReadConfig() {

	// Open the XML file
	TiXmlDocument Document(Save::Instance().GetConfigFile().string().c_str());
	if(!Document.LoadFile()) {
		printf("line %d column %d: %s\n", Document.ErrorRow(), Document.ErrorCol(), Document.ErrorDesc());
		return 0;
	}

	// Check for config tag
	TiXmlElement *ConfigElement = Document.FirstChildElement("config");
	if(!ConfigElement)
		return 0;

	// Check for a video tag
	TiXmlElement *VideoElement = ConfigElement->FirstChildElement("video");
	if(VideoElement) {

		// Get driver
		VideoElement->QueryIntAttribute("driver", (int *)(&DriverType));

		// Check for the screen tag
		TiXmlElement *ScreenElement = VideoElement->FirstChildElement("screen");
		if(ScreenElement) {

			// Get screen attributes
			ScreenElement->QueryIntAttribute("width", &ScreenWidth);
			ScreenElement->QueryIntAttribute("height", &ScreenHeight);
			ScreenElement->QueryIntAttribute("fullscreen", &FullScreen);

		}
	}

	// Check for the audio tag
	TiXmlElement *AudioElement = ConfigElement->FirstChildElement("audio");
	if(AudioElement) {
		
		// Get audio attributes
		AudioElement->QueryIntAttribute("enabled", &AudioEnabled);
			
		// Get sound element
		TiXmlElement *SoundElement = AudioElement->FirstChildElement("sound");
		if(!SoundElement) {

			// Get sound attributes
			SoundElement->QueryFloatAttribute("volume", &SoundVolume);
		}

		// Get music element
		TiXmlElement *MusicElement = AudioElement->FirstChildElement("music");
		if(!MusicElement) {

			// Get music attributes
			MusicElement->QueryFloatAttribute("volume", &MusicVolume);
		}
	}

	// Get input element
	TiXmlElement *InputElement = ConfigElement->FirstChildElement("input");
	if(InputElement) {

		// Get mouse attributes
		InputElement->QueryFloatAttribute("mousex", &MouseScaleX);

		// Get mouse attributes
		InputElement->QueryFloatAttribute("mousey", &MouseScaleY);
	}

	// Get keyboard mapping
	for(TiXmlElement *ActionElement = InputElement->FirstChildElement("action"); ActionElement != 0; ActionElement = ActionElement->NextSiblingElement("action")) {
	
		// Get action type
		int ActionType;
		if(ActionElement->QueryIntAttribute("type", &ActionType) != TIXML_SUCCESS)
			continue;

		// Get key
		int KeyCode;
		if(ActionElement->QueryIntAttribute("key", &KeyCode) != TIXML_SUCCESS)
			continue;

		// Assign key
		if(ActionType >= 0 && ActionType < ACTIONS::COUNT && KeyCode >= 0 && KeyCode < KEY_KEY_CODES_COUNT)
			Keys[ActionType] = (EKEY_CODE)KeyCode;
		
	}

	// Replays
	TiXmlElement *ReplayElement = ConfigElement->FirstChildElement("replay");
	if(ReplayElement) {
		ReplayElement->QueryIntAttribute("interval", (int *)(&ReplayInterval));
	}

	return 1;
}

// Writes the config file
int ConfigClass::WriteConfig() {

	TiXmlDocument Document;

	// Create header
	Document.LinkEndChild(new TiXmlDeclaration("1.0", "", ""));
	
	// Create config element
	TiXmlElement *ConfigElement = new TiXmlElement("config");
	ConfigElement->SetAttribute("version", "1.0");
	Document.LinkEndChild(ConfigElement);

	// Create video element
	TiXmlElement *VideoElement = new TiXmlElement("video");
	VideoElement->SetAttribute("driver", DriverType);
	ConfigElement->LinkEndChild(VideoElement);

	// Screen settings
	TiXmlElement *ScreenElement = new TiXmlElement("screen");
	ScreenElement->SetAttribute("width", ScreenWidth);
	ScreenElement->SetAttribute("height", ScreenHeight);
	ScreenElement->SetAttribute("fullscreen", FullScreen);
	VideoElement->LinkEndChild(ScreenElement);

	// Create audio element
	TiXmlElement *AudioElement = new TiXmlElement("audio");
	AudioElement->SetAttribute("enabled", AudioEnabled);
	ConfigElement->LinkEndChild(AudioElement);

	// Sound
	TiXmlElement *SoundElement = new TiXmlElement("sound");
	SoundElement->SetDoubleAttribute("volume", SoundVolume);
	AudioElement->LinkEndChild(SoundElement);

	// Music
	TiXmlElement *MusicElement = new TiXmlElement("music");
	MusicElement->SetDoubleAttribute("volume", MusicVolume);
	AudioElement->LinkEndChild(MusicElement);

	// Input
	TiXmlElement *InputElement = new TiXmlElement("input");
	InputElement->SetDoubleAttribute("mousex", 1.0);
	InputElement->SetDoubleAttribute("mousey", 1.0);
	ConfigElement->LinkEndChild(InputElement);

	// Actions
	for(int i = 0; i < ACTIONS::COUNT; i++) {
		TiXmlElement *ActionElement = new TiXmlElement("action");
		ActionElement->SetAttribute("type", i);
		ActionElement->SetAttribute("key", Keys[i]);
		InputElement->LinkEndChild(ActionElement);
	}

	// Replays
	TiXmlElement *ReplayElement = new TiXmlElement("replay");
	ReplayElement->SetAttribute("interval", ReplayInterval);
	ConfigElement->LinkEndChild(ReplayElement);

	// Write file
	Document.SaveFile(Save::Instance().GetConfigFile().string().c_str());

	return 1;
}
