/*************************************************************************************
*	irrlamb - http://code.google.com/p/irrlamb/
*	Copyright (C) 2007  Alan Witkowski
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation, either version 3 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program.  If not, see <http://www.gnu.org/licenses/>.
**************************************************************************************/
#include "game.h"
#include "input.h"
#include "globals.h"
#include "graphics.h"
#include "audio.h"
#include "interface.h"
#include "state.h"
#include "random.h"
#include "scripting.h"
#include "objectmanager.h"
#include "config.h"
#include "save.h"
#include "campaign.h"
#include "../play.h"
#include "../viewreplay.h"
#include "../menu.h"

#ifdef _WIN32
	#include <winbase.h>
#else
	#include <unistd.h>
#endif

// Constructor
GameClass::GameClass()
:	NewState(NULL),
	ChangeType(NONE),
	Done(false),
	MouseWasLocked(false),
	MouseWasShown(true),
	PreviousWindowActive(1),
	WindowActive(1),
	QuitOnEscape(false),
	TimeStamp(0) {

}

// Processes parameters and initializes the game
int GameClass::Init(int TCount, char **TArguments) {

	// Defaults
	StateClass *FirstState = MenuState::Instance();
	bool AudioEnabled = true;

	// Process arguments
	stringc Token;
	int TokensRemaining;
	for(int i = 1; i < TCount; i++) {
		Token = stringc(TArguments[i]);
		TokensRemaining = TCount - i - 1;
		if(Token == "-level" && TokensRemaining > 0) {
			PlayState::Instance()->SetNextLevel(TArguments[++i]);
			FirstState = PlayState::Instance();
			QuitOnEscape = true;
		}
		else if(Token == "-replay" && TokensRemaining > 0) {
			ViewReplayState::Instance()->SetCurrentReplay(TArguments[++i]);
			FirstState = ViewReplayState::Instance();
			QuitOnEscape = true;
		}
		else if(Token == "-noaudio") {
			AudioEnabled = false;
		}
	}

	// Set up the save system
	if(!Save::Instance().Init())
		return 0;

	// Set up config system
	if(!Config::Instance().Init())
		return 0;

	// Read the config file
	int HasConfigFile = Config::Instance().ReadConfig();

	// Set up the graphics
	if(!Graphics::Instance().Init(Config::Instance().ScreenWidth, Config::Instance().ScreenHeight, Config::Instance().FullScreen, Config::Instance().DriverType, &Input::Instance()))
		return 0;

	// Write a config file if none exists
	if(!HasConfigFile)
		Config::Instance().WriteConfig();

	// Set random seed
	Random::Instance().SetSeed(irrTimer->getRealTime());

	// Set up the audio system
	if(!Audio::Instance().Init(AudioEnabled && Config::Instance().AudioEnabled))
		return 0;

	// Initializes replay system
	Save::Instance().FindHighestReplayNumber();

	// Set up the interface system
	if(!Interface::Instance().Init())
		return 0;

	// Set up the scripting system
	if(!Scripting::Instance().Init())
		return 0;

	// Set up the object manager
	if(!ObjectManager::Instance().Init())
		return 0;

	// Set up the object manager
	if(!Campaign::Instance().Init())
		return 0;

	// Load stats file
	Save::Instance().LoadLevelStats();

	// Start the state timer
	ResetTimer();

	// Set the first state
	if(!ChangeState(FirstState)) {
		irrLog->log("Failed to set the first state");
		return 0;
	}
	
	return 1;
}

// Attempts to change the state. Returns 1 for success, 0 for fail
int GameClass::ChangeState(StateClass *TState) {

	// If no other states exist, initialize it immediately
	if(States.empty()) {
		ResetGraphics();	

		// Attempt to initialize the new state
		if(!TState->Init())
			return 0;

		ResetTimer();
		States.push(TState);
		return 1;
	}

	// Set the new state so the manager can initialize it later
	NewState = TState;
	ChangeType = CHANGE;
	
	return 1;
}

// Pop the top of the state stack
int GameClass::PopState() {

	// Check size
	if(States.empty())
		return 0;

	// Close last state
	int Result = States.top()->Close();
	States.pop();

	return Result;
}

// Updates the current state and manages the state stack
void GameClass::Update() {

	// Run irrlicht engine
	if(!irrDevice->run())
		Done = true;

	// Get time difference from last frame
	DeltaTime = irrTimer->getTime() - TimeStamp;
	TimeStamp = irrTimer->getTime();
	
	// Limit frame rate
	if(DeltaTime <= 12)
		Delay(12 - DeltaTime);

	// Check for window activity
	PreviousWindowActive = WindowActive;
	WindowActive = irrDevice->isWindowActive();

	// Check for window focus/blur events
	if(PreviousWindowActive != WindowActive) {
		Input::Instance().ResetInputState();

		if(!WindowActive) {
			MouseWasLocked = Input::Instance().GetMouseLocked();
			MouseWasShown = irrDevice->getCursorControl()->isVisible();
			Input::Instance().SetMouseLocked(false);
			irrDevice->getCursorControl()->setVisible(true);
		}			
		else {
			Input::Instance().SetMouseLocked(MouseWasLocked);
			irrDevice->getCursorControl()->setVisible(MouseWasShown);
		}

	}

	// Update the current state
	if(!States.empty()) {

		// Update
		States.top()->Update(DeltaTime);
		
		// Draw
		Graphics::Instance().BeginFrame();
		States.top()->Draw();
		Graphics::Instance().EndFrame();
	}

	// Handle the new state request
	if(NewState != NULL) {
		ResetGraphics();

		// Handle the request type
		switch(ChangeType) {
			case CHANGE:
				
				// Pop previous state
				if(!PopState()) {
					irrLog->log("Failed to pop previous state");
					Done = true;
				}

				// Attempt to initialize the new state
				if(NewState->Init()) {
					States.push(NewState);
				}
				else {
					irrLog->log("Failed to initialize new state");
					ChangeState(MenuState::Instance());
				}
				
			break;
			case PUSH:

				// Attempt to initialize the new state
				if(NewState->Init())
					States.push(NewState);
			break;
			default:
				// Bad type
			break;
		}

		ResetTimer();

		NewState = NULL;
		ChangeType = NONE;
	}

}

// Pop all the states off and shuts down the system
void GameClass::Close() {
	
	// Close all the states
	while (!States.empty()) {

		if(!States.top()->Close()) {
			// Error
		}

		States.pop();
	}

	Config::Instance().WriteConfig();

	// Shut down the system
	Campaign::Instance().Close();
	ObjectManager::Instance().Close();
	Scripting::Instance().Close();
	Interface::Instance().Close();
	Audio::Instance().Close();
	Graphics::Instance().Close();
	Config::Instance().Close();
	Save::Instance().Close();
}

// Returns the current game state
StateClass *GameClass::GetCurrentState() {

	return States.top();
}

// Resets the game timer
void GameClass::ResetTimer() {
	
	irrTimer->setTime(0);
	TimeStamp = irrTimer->getTime();
}

// Resets the graphics for a state
void GameClass::ResetGraphics() {
	Graphics::Instance().SetClearColor(SColor(0, 0, 0, 0));
	Graphics::Instance().SetDrawScene(true);
	Interface::Instance().Clear();
}

// Delays execution of the program
void GameClass::Delay(u32 TTime) {
	#ifdef _WIN32
		Sleep(TTime);
	#else
		usleep(TTime);
	#endif
}
