/*************************************************************************************
*	irrlamb - http://code.google.com/p/irrlamb/
*	Copyright (C) 2007  Alan Witkowski
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation, either version 3 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program.  If not, see <http://www.gnu.org/licenses/>.
**************************************************************************************/
#include "level.h"
#include "globals.h"
#include "graphics.h"
#include "objectmanager.h"
#include "physics.h"
#include "scripting.h"
#include "replay.h"
#include "../tinyxml/tinyxml.h"
#include "../objects/default.h"
#include "../objects/scene.h"
#include "../objects/player.h"
#include "../objects/sphere.h"
#include "../objects/box.h"
#include "../objects/mesh.h"
#include "../objects/orb.h"
#include "../objects/terrain.h"
#include "../objects/water.h"
#include "../objects/zone.h"
#include "../objects/basicjoint.h"
#include "../objects/springjoint.h"
#include "../objects/grapple.h"

// Loads a level file
int LevelClass::Init(const stringc &TFile, bool TRecord) {
	ClearColor.set(255, 0, 0, 0);
	DefaultPosition.set(0.0f, 0.0f, 0.0f);
	DefaultRotation.set(0.0f, 0.0f, 0.0f);
	for(int i = 0; i < CUSTOM_COUNT; i++)
		CustomTemplates[i] = -1;

	// Open the XML file
	stringc LevelFile = stringc("levels/") + TFile;
	TiXmlDocument Document(LevelFile.c_str());
	if(!Document.LoadFile()) {
		stringc Buffer = stringc("line ") + stringc(Document.ErrorRow()) + stringc(" column ")
						+ stringc(Document.ErrorCol()) + stringc(": ") + stringc(Document.ErrorDesc());
		irrLog->log(Buffer.c_str());
		Close();
		return 0;
	}

	// Check for level tag
	TiXmlElement *LevelElement = Document.FirstChildElement("level");
	if(!LevelElement) {
		irrLog->log("Could not find level tag");
		Close();
		return 0;
	}

	// Check level id
	LevelID = LevelElement->Attribute("id");
	if(LevelID == "") {
		irrLog->log("Could not find level id");
		Close();
		return 0;
	}

	// Check level version
	if(LevelElement->QueryIntAttribute("version", &LevelVersion) != TIXML_SUCCESS) {
		irrLog->log("Could not find level version");
		Close();
		return 0;
	}

	// Load level information
	TiXmlElement *InfoElement = LevelElement->FirstChildElement("info");

	// Load resources
	TiXmlElement *ResourcesElement = LevelElement->FirstChildElement("resources");

	// Graphics
	TiXmlElement *GraphicsElement = LevelElement->FirstChildElement("graphics");
	if(GraphicsElement) {
		TiXmlElement *Element;
		
		// Clear color
		Element = GraphicsElement->FirstChildElement("clear");
		if(Element) {
			int Colors[3] = { ClearColor.getRed(), ClearColor.getGreen(), ClearColor.getBlue() };
			Element->QueryIntAttribute("r", &Colors[0]);
			Element->QueryIntAttribute("g", &Colors[1]);
			Element->QueryIntAttribute("b", &Colors[2]);

			ClearColor.set(255, Colors[0], Colors[1], Colors[2]);
		}

		// Fog
		Element = GraphicsElement->FirstChildElement("fog");
		if(Element) {
			int Colors[3] = { 0, 0, 0 };
			float Attributes[3] = { 10.0f, 50.0f, 0.05f };
			int Linear = 1;

			Element->QueryIntAttribute("r", &Colors[0]);
			Element->QueryIntAttribute("g", &Colors[1]);
			Element->QueryIntAttribute("b", &Colors[2]);
			Element->QueryFloatAttribute("start", &Attributes[0]);
			Element->QueryFloatAttribute("end", &Attributes[1]);
			Element->QueryFloatAttribute("density", &Attributes[2]);
			Element->QueryIntAttribute("linear", &Linear);

			irrDriver->setFog(SColor(255, Colors[0], Colors[1], Colors[2]), Linear, Attributes[0], Attributes[1], Attributes[2], true);
		}
	}

	// Get scripts
	array<stringc> Scripts;
	TiXmlElement *ScriptElement = ResourcesElement->FirstChildElement("script");
	for(; ScriptElement != 0; ScriptElement = ScriptElement->NextSiblingElement("script"))
		Scripts.push_back(ScriptElement->GetText());

	// Load objects
	if(!LoadObjects(LevelElement->FirstChildElement("objects"), false)) {
		Close();
		return 0;
	}

	// Templates
	if(!LoadObjects(LevelElement->FirstChildElement("templates"), true)) {
		Close();
		return 0;
	}

	AddBuiltInTemplates();

	// Run scripts
	if(TRecord) {
		Replay::Instance().StartRecording(TFile.c_str(), LevelVersion);

		// Run scripts
		Scripting::Instance().Reset();
		for(u32 i = 0; i < Scripts.size(); i++) {
			
			// Load a level
			if(!Scripting::Instance().LoadFile(Scripts[i]))
				return 0;
		}
	}

	SetClearColor();

	return 1;
}

// Closes the level system
int LevelClass::Close() {

	ObjectManager::Instance().ClearObjects();
	for(u32 i = 0; i < Templates.size(); i++)
		delete Templates[i];

	Templates.clear();

	return 1;
}

// Loads all the objects in a tag
int LevelClass::LoadObjects(TiXmlElement *TElement, bool TTemplate) {

	if(TElement) {

		for(TiXmlElement *ObjectElement = TElement->FirstChildElement(); ObjectElement != 0; ObjectElement = ObjectElement->NextSiblingElement()) {
			
			// Create a template
			DefaultObjectStruct *Object = new DefaultObjectStruct;

			// Get the object properties
			if(!GetObjectProperties(ObjectElement, *Object))
				return 0;

			// Save for Lua
			if(TTemplate)
				Templates.push_back(Object);
			else
				CreateObject(*Object);
		}
	}

	return 1;
}

// Processes an object tag
int LevelClass::GetObjectProperties(TiXmlElement *TObjectElement, DefaultObjectStruct &TObject) {

	TiXmlElement *Element;

	// Get type
	stringc ObjectType = TObjectElement->Value();

	// Get name
	TObject.Name = TObjectElement->Attribute("name");
	if(TObject.Name == "") {
		irrLog->log("Object is missing name");
		return 0;
	}
	
	// Get visible
	TObjectElement->QueryIntAttribute("visible", &TObject.Visible);

	// Get lifetime
	TObjectElement->QueryFloatAttribute("lifetime", &TObject.Lifetime);

	// Get position
	Element = TObjectElement->FirstChildElement("position");
	if(Element) {
		Element->QueryFloatAttribute("x", &TObject.Position.X);
		Element->QueryFloatAttribute("y", &TObject.Position.Y);
		Element->QueryFloatAttribute("z", &TObject.Position.Z);
	}

	// Get euler rotation
	Element = TObjectElement->FirstChildElement("rotation");
	if(Element) {
		Element->QueryFloatAttribute("x", &TObject.Rotation.X);
		Element->QueryFloatAttribute("y", &TObject.Rotation.Y);
		Element->QueryFloatAttribute("z", &TObject.Rotation.Z);
	}

	// Get scale
	Element = TObjectElement->FirstChildElement("scale");
	if(Element) {
		Element->QueryFloatAttribute("x", &TObject.Scale.X);
		Element->QueryFloatAttribute("y", &TObject.Scale.Y);
		Element->QueryFloatAttribute("z", &TObject.Scale.Z);
	}

	// Get radius
	Element = TObjectElement->FirstChildElement("radius");
	if(Element)
		TObject.Radius = (float)atof(Element->GetText());

	// Get physical attributes
	Element = TObjectElement->FirstChildElement("physics");
	if(Element) {
		Element->QueryFloatAttribute("mass", &TObject.Mass);
		Element->QueryFloatAttribute("friction", &TObject.Friction);
	}

	// Get damping
	Element = TObjectElement->FirstChildElement("damping");
	if(Element) {
		Element->QueryFloatAttribute("linear", &TObject.LinearDamping);
		Element->QueryFloatAttribute("angular", &TObject.AngularDamping);
	}

	// Get collision information
	Element = TObjectElement->FirstChildElement("collision");
	if(Element) {
		int CameraCollide = 1;
		TObject.CollisionCallback = Element->Attribute("callback");
		Element->QueryIntAttribute("camera", &CameraCollide);

		if(!CameraCollide)
			Physics::Instance().RemoveFilter(TObject.CollisionGroup, PhysicsClass::FILTER_CAMERA);
	}

	// Get file
	Element = TObjectElement->FirstChildElement("file");
	if(Element)
		TObject.File = Element->GetText();

	// Get texture
	Element = TObjectElement->FirstChildElement("texture");
	if(Element) {
		stringc Attribute;
		
		// Get normal texture
		Attribute = Element->Attribute("normal");
		if(Attribute != "") {
			TObject.Texture = stringc("textures/") + Attribute;
			if(!irrFile->existFile(TObject.Texture.c_str())) {
				irrLog->log("Texture file does not exist");
				return 0;
			}
		}

		// Get detail texture
		Attribute = Element->Attribute("detail");
		if(Attribute != "") {
			TObject.DetailTexture = stringc("textures/") + Attribute;
			if(!irrFile->existFile(TObject.DetailTexture.c_str())) {
				irrLog->log("Detail texture file does not exist");
				return 0;
			}
		}
	}

	// Get material properties
	Element = TObjectElement->FirstChildElement("material");
	if(Element) {
		Element->QueryIntAttribute("lighting", &TObject.Lighting);
		Element->QueryIntAttribute("fog", &TObject.Fog);
	}

	// Validate objects
	if(ObjectType == "scene") {
		TObject.File = stringc("scenes/") + TObject.File;
		if(!irrFile->existFile(TObject.File.c_str())) {
			irrLog->log("Scene file does not exist");
			return 0;
		}

		TObject.Mass = 0.0f;
		TObject.Type = ObjectClass::SCENE;
	}
	else if(ObjectType == "player") {
		TObjectElement->QueryFloatAttribute("torque", &TObject.TorqueFactor);
		TObjectElement->QueryIntAttribute("grapple", &TObject.GrappleEnabled);
		Physics::Instance().RemoveFilter(TObject.CollisionMask, PhysicsClass::FILTER_GRAPPLE);

		TObject.Type = ObjectClass::PLAYER;
	}
	else if(ObjectType == "sphere") {
		TObject.Type = ObjectClass::SPHERE;
	}
	else if(ObjectType == "box") {
		TObject.Type = ObjectClass::BOX;
	}
	else if(ObjectType == "mesh") {
		TObject.File = stringc("meshes/") + TObject.File;
		if(!irrFile->existFile(TObject.File.c_str())) {
			irrLog->log("Mesh file does not exist");
			return 0;
		}

		TObject.Mass = 0.0f;
		TObject.Type = ObjectClass::MESH;
	}
	else if(ObjectType == "orb") {
		TObject.Type = ObjectClass::ORB;
	}
	else if(ObjectType == "terrain") {
		TObject.File = stringc("terrain/") + TObject.File;
		if(!irrFile->existFile(TObject.File.c_str())) {
			irrLog->log("Terrain file does not exist");
			return 0;
		}

		TObjectElement->QueryIntAttribute("smooth", &TObject.SmoothFactor);

		TObject.Mass = 0.0f;
		TObject.Type = ObjectClass::TERRAIN;
	}
	else if(ObjectType == "water") {
		TObject.Mass = 0.0f;
		TObject.Type = ObjectClass::WATER;
	}
	else if(ObjectType == "zone") {
		TObjectElement->QueryIntAttribute("active", &TObject.Active);

		TObject.Type = ObjectClass::ZONE;
	}
	else if(ObjectType == "basicjoint") {

		// First body
		Element = TObjectElement->FirstChildElement("body1");
		if(Element)
			TObject.BodyA = static_cast<RigidBodyClass *>(ObjectManager::Instance().GetObjectByName(Element->GetText()));

		// Second body
		Element = TObjectElement->FirstChildElement("body2");
		if(Element)
			TObject.BodyB = static_cast<RigidBodyClass *>(ObjectManager::Instance().GetObjectByName(Element->GetText()));

		// Linear limit
		for(Element = TObjectElement->FirstChildElement("linearlimit"); Element != 0; Element = Element->NextSiblingElement("linearlimit")) {
			int Type = 0;
			Element->QueryIntAttribute("type", &Type);
			Element->QueryFloatAttribute("x", &TObject.LinearLimit[Type][0]);
			Element->QueryFloatAttribute("y", &TObject.LinearLimit[Type][1]);
			Element->QueryFloatAttribute("z", &TObject.LinearLimit[Type][2]);
		}

		// Angular limit
		for(Element = TObjectElement->FirstChildElement("angularlimit"); Element != 0; Element = Element->NextSiblingElement("angularlimit")) {
			int Type = 0;
			Element->QueryIntAttribute("type", &Type);
			Element->QueryFloatAttribute("x", &TObject.AngularLimit[Type][0]);
			Element->QueryFloatAttribute("y", &TObject.AngularLimit[Type][1]);
			Element->QueryFloatAttribute("z", &TObject.AngularLimit[Type][2]);
		}

		TObject.Type = ObjectClass::BASICJOINT;
	}
	else if(ObjectType == "springjoint") {

		// Properties
		TObjectElement->QueryFloatAttribute("distance", &TObject.SpringDistance);
		TObjectElement->QueryFloatAttribute("constant", &TObject.SpringConstant);

		// First body
		Element = TObjectElement->FirstChildElement("body1");
		if(Element)
			TObject.BodyA = static_cast<RigidBodyClass *>(ObjectManager::Instance().GetObjectByName(Element->GetText()));

		// Second body
		Element = TObjectElement->FirstChildElement("body2");
		if(Element)
			TObject.BodyB = static_cast<RigidBodyClass *>(ObjectManager::Instance().GetObjectByName(Element->GetText()));

		TObject.Type = ObjectClass::SPRINGJOINT;
	}

	return 1;
}

// Creates an object from a template
ObjectClass *LevelClass::CreateObject(DefaultObjectStruct &TObject, bool TUseTemplateOrientation) {
	
	// Use the level's orientation
	if(!TUseTemplateOrientation) {
		TObject.Position = DefaultPosition;
		TObject.Rotation = DefaultRotation;
	}

	// Create object
	switch(TObject.Type) {
		case ObjectClass::SCENE:
			return ObjectManager::Instance().AddObject(new SceneClass(TObject));
		break;
		case ObjectClass::PLAYER:
			return ObjectManager::Instance().AddObject(new PlayerClass(TObject));
		break;
		case ObjectClass::SPHERE:
			return ObjectManager::Instance().AddObject(new SphereClass(TObject));
		break;
		case ObjectClass::BOX:
			return ObjectManager::Instance().AddObject(new BoxClass(TObject));
		break;
		case ObjectClass::MESH:
			return ObjectManager::Instance().AddObject(new MeshClass(TObject));
		break;
		case ObjectClass::ORB:
			return ObjectManager::Instance().AddObject(new OrbClass(TObject));
		break;
		case ObjectClass::TERRAIN:
			return ObjectManager::Instance().AddObject(new TerrainClass(TObject));
		break;
		case ObjectClass::WATER:
			return ObjectManager::Instance().AddObject(new WaterClass(TObject));
		break;
		case ObjectClass::ZONE:
			return ObjectManager::Instance().AddObject(new ZoneClass(TObject));
		break;
		case ObjectClass::BASICJOINT:
			return ObjectManager::Instance().AddObject(new BasicJointClass(TObject));
		break;
		case ObjectClass::SPRINGJOINT:
			return ObjectManager::Instance().AddObject(new SpringJointClass(TObject));
		break;
		case ObjectClass::GRAPPLE:
			return ObjectManager::Instance().AddObject(new GrappleClass(TObject));
		break;
	}

	return NULL;
}

// Creates an object from a template index
ObjectClass *LevelClass::CreateObjectFromTemplateIndex(int TIndex) {
	
	if(TIndex < 0 || TIndex >= (int)Templates.size()) {
		printf("Bad Template Index: %d\n", TIndex);
		return NULL;
	}

	WriteCreateEvent(TIndex);

	return CreateObject(*Templates[TIndex], false);
}

// Writes an object create event to the replay file
void LevelClass::WriteCreateEvent(int TIndex) {

	// Start writing replay events after the initial level load
	if(Replay::Instance().IsRecording()) {
		Replay::Instance().WriteEvent(ReplayClass::PACKET_CREATE);

		ReplayObjectCreateStruct Packet;
		Packet.Template = TIndex;
		Packet.Position = DefaultPosition;
		Packet.Rotation = DefaultRotation;
		Replay::Instance().WritePacket(Packet);
	}
}

// Returns the index of a template given a name, -1 is returned when no template is found
int LevelClass::GetTemplateFromName(const stringc &TName) {

	// Search for the name
	for(u32 i = 0; i < Templates.size(); i++) {
		if(Templates[i]->Name == TName)
			return i;
	}

	return -1;
}

// Returns an object struct from a template index
bool LevelClass::GetTemplateFromIndex(int TIndex, DefaultObjectStruct &TObject) {
	
	if(TIndex < 0 || TIndex >= (int)Templates.size())
		return false;

	TObject = *Templates[TIndex];

	return true;
}

// Sets the level's clear color
void LevelClass::SetClearColor() {
	
	Graphics::Instance().SetClearColor(ClearColor);
}

// Adds custom object templates to the level
void LevelClass::AddBuiltInTemplates() {

	DefaultObjectStruct *GrappleTip = new DefaultObjectStruct;
	GrappleTip->Radius = 0.07f;
	GrappleTip->Type = ObjectClass::GRAPPLE;
	GrappleTip->Name = "_grapple";
	GrappleTip->Texture = "textures/orb0.png";
	GrappleTip->Mass = 0.1f;
	GrappleTip->LinearDamping = 0.5f;
	GrappleTip->Hookable = false;
	Physics::Instance().RemoveFilter(GrappleTip->CollisionGroup, PhysicsClass::FILTER_CAMERA);
	Templates.push_back(GrappleTip);
	CustomTemplates[CUSTOM_GRAPPLETIP] = Templates.size() - 1;

	DefaultObjectStruct *GrappleFixed = new DefaultObjectStruct;
	GrappleFixed->Name = "_grapplefixedjoint";
	GrappleFixed->Type = ObjectClass::BASICJOINT;
	Templates.push_back(GrappleFixed);
	CustomTemplates[CUSTOM_GRAPPLEFIXED] = Templates.size() - 1;

	DefaultObjectStruct *GrappleSpring = new DefaultObjectStruct;
	GrappleSpring->Name = "_grapplespringjoint";
	GrappleSpring->Type = ObjectClass::SPRINGJOINT;
	GrappleSpring->SpringConstant = 2.0f;
	Templates.push_back(GrappleSpring);
	CustomTemplates[CUSTOM_GRAPPLESPRING] = Templates.size() - 1;

}
