/*************************************************************************************
*	irrlamb - http://code.google.com/p/irrlamb/
*	Copyright (C) 2007  Alan Witkowski
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation, either version 3 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program.  If not, see <http://www.gnu.org/licenses/>.
**************************************************************************************/
#include "objectmanager.h"
#include "replay.h"
#include "../objects/object.h"
#include "../objects/rigidbody.h"

// Constructor
ObjectManagerClass::ObjectManagerClass() {

}

// Initializes the level manager
int ObjectManagerClass::Init() {

	NextObjectID = 0;

	return 1;
}

// Closes the graphics system
int ObjectManagerClass::Close() {

	ClearObjects();

	return 1;
}

// Adds an object to the manager
ObjectClass *ObjectManagerClass::AddObject(ObjectClass *TObject) {

	if(TObject != NULL) {
		
		// Set replay ID
		TObject->SetID(NextObjectID);
		NextObjectID++;

		Objects.push_back(TObject);
	}

	return TObject;
}

// Deletes an object
void ObjectManagerClass::DeleteObject(ObjectClass *TObject) {

	TObject->SetDeleted(true);
}

// Gets an object by name
ObjectClass *ObjectManagerClass::GetObjectByName(const stringc &TName) {

	// Search through object list
	for(list<ObjectClass *>::Iterator Iterator = Objects.begin(); Iterator != Objects.end(); ++Iterator) {
		if((*Iterator)->GetName() == TName)
			return *Iterator;
	}

	return NULL;
}

// Deletes all of the objects
void ObjectManagerClass::ClearObjects() {
	
	// Delete constraints first
	for(list<ObjectClass *>::Iterator Iterator = Objects.begin(); Iterator != Objects.end(); ) {
		if((*Iterator)->GetType() == ObjectClass::BASICJOINT) {
			delete (*Iterator);
			(*Iterator) = NULL;
			Iterator = Objects.erase(Iterator);
		}
		else
			++Iterator;
	}

	// Delete objects
	for(list<ObjectClass *>::Iterator Iterator = Objects.begin(); Iterator != Objects.end(); ++Iterator) {
		delete (*Iterator);
	}

	Objects.clear();
	NextObjectID = 0;
}

// Performs start frame operations on the objects
void ObjectManagerClass::BeginFrame() {
	
	for(list<ObjectClass *>::Iterator Iterator = Objects.begin(); Iterator != Objects.end(); ++Iterator)
		(*Iterator)->BeginFrame();

}

// Performs end frame operations on the objects
void ObjectManagerClass::EndFrame() {
	
	for(list<ObjectClass *>::Iterator Iterator = Objects.begin(); Iterator != Objects.end(); ++Iterator)
		(*Iterator)->EndFrame();

}

// Updates all objects in the scene
void ObjectManagerClass::Update(u32 TDeltaTime) {

	bool UpdateReplay = Replay::Instance().NeedsPacket();
	int ReplayMovementCount = 0;

	// Update objects
	for(list<ObjectClass *>::Iterator Iterator = Objects.begin(); Iterator != Objects.end(); ) {

		// Update the object
		(*Iterator)->Update(TDeltaTime);

		// Delete old objects
		if((*Iterator)->GetDeleted()) {

			// Write delete events to the replay
			if(Replay::Instance().IsRecording()) {
				Replay::Instance().WriteEvent(ReplayClass::PACKET_DELETE);

				ReplayObjectDeleteStruct Packet;
				Packet.ObjectID = (*Iterator)->GetID();
				Replay::Instance().WritePacket(Packet);
			}

			delete (*Iterator);
			Iterator = Objects.erase(Iterator);
		}
		else {
			
			// Update the object's graphics
			(*Iterator)->UpdateRender();
			if((*Iterator)->HasMovementChanged())
				ReplayMovementCount++;

			++Iterator;
		}
	}

	// Write a replay movement packet
	if(UpdateReplay && ReplayMovementCount > 0) {

		// Write the event
		Replay::Instance().WriteEvent(ReplayClass::PACKET_MOVEMENT);

		// Write the header
		ReplayObjectHeaderStruct HeaderPacket;
		HeaderPacket.Count = ReplayMovementCount;
		Replay::Instance().WritePacket(HeaderPacket);

		// Write the updated objects
		ReplayMovementStruct ObjectPacket;
		int ObjectIndex = 0;
		for(list<ObjectClass *>::Iterator Iterator = Objects.begin(); Iterator != Objects.end(); ++Iterator) {

			// Save the replay
			if((*Iterator)->GetReplayPacket(&ObjectPacket)) {
				ObjectPacket.ObjectIndex = ObjectIndex;
				Replay::Instance().WritePacket(ObjectPacket);
			}
			ObjectIndex++;
		}

		//printf("ObjectIndex=%d\n", ObjectIndex);
	}
}

// Update special replays function for each object
void ObjectManagerClass::UpdateReplay(u32 TDeltaTime) {

	// Update objects
	for(list<ObjectClass *>::Iterator Iterator = Objects.begin(); Iterator != Objects.end(); ++Iterator)
		(*Iterator)->UpdateReplay(TDeltaTime);
}

// Updates all objects in the scene from a replay file
void ObjectManagerClass::UpdateFromMovementPacket(const ReplayEventStruct &TPacket) {

	// Read object header
	ReplayObjectHeaderStruct HeaderPacket;
	Replay::Instance().ReadPacket(HeaderPacket);

	// Read first object packet
	ReplayMovementStruct ObjectPacket;
	Replay::Instance().ReadPacket(ObjectPacket);

	// Loop through the rest of the objects
	int ObjectIndex = 0, UpdatedObjectCount = 0;
	for(list<ObjectClass *>::Iterator Iterator = Objects.begin(); Iterator != Objects.end(); ++Iterator) {
		if(ObjectPacket.ObjectIndex == ObjectIndex) {

			RigidBodyClass *RigidBody = static_cast<RigidBodyClass *>(*Iterator);
			RigidBody->GetNode()->setPosition(ObjectPacket.Position);
			RigidBody->GetNode()->setRotation(ObjectPacket.Rotation);

			if(UpdatedObjectCount < HeaderPacket.Count - 1)
				Replay::Instance().ReadPacket(ObjectPacket);

			UpdatedObjectCount++;
		}
		ObjectIndex++;
	}

	//printf("ObjectIndex=%d\n", ObjectIndex);
}

// Returns an object by an index, NULL if no such index
ObjectClass *ObjectManagerClass::GetObjectByID(int TID) {
	
	for(list<ObjectClass *>::Iterator Iterator = Objects.begin(); Iterator != Objects.end(); ++Iterator) {
		if((*Iterator)->GetID() == TID)
			return *Iterator;

	}

	return NULL;
}

// Deletes an object by its ID
void ObjectManagerClass::DeleteObjectByID(int TID) {
	
	for(list<ObjectClass *>::Iterator Iterator = Objects.begin(); Iterator != Objects.end(); ++Iterator) {
		if((*Iterator)->GetID() == TID) {
			delete (*Iterator);
			Objects.erase(Iterator);
			return;
		}
	}

	printf("Bad Object ID: %d\n", TID);
}

