/*************************************************************************************
*	irrlamb - http://code.google.com/p/irrlamb/
*	Copyright (C) 2007  Alan Witkowski
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation, either version 3 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program.  If not, see <http://www.gnu.org/licenses/>.
**************************************************************************************/
#include "physics.h"
#include "globals.h"
#include "../objects/object.h"
#include "../engine/game.h"

// Constructor
PhysicsClass::PhysicsClass()
:	Dispatcher(NULL),
	BroadPhase(NULL),
	Solver(NULL) {

}

// Initializes the physics system
int PhysicsClass::Init(bool TEnabled) {

	Enabled = TEnabled;
	if(Enabled) {
		SimulationSpeed = 1.0f;

		// Set up physics modules
		CollisionConfiguration = new btDefaultCollisionConfiguration();
		BroadPhase = new btAxisSweep3(btVector3(-1000, -1000, -1000), btVector3(1000, 1000, 1000), 32766);
		Dispatcher = new btCollisionDispatcher(CollisionConfiguration);
		Solver = new btSequentialImpulseConstraintSolver();
		World = new btDiscreteDynamicsWorld(Dispatcher, BroadPhase, Solver, CollisionConfiguration);
		World->setGravity(btVector3(0.0f, -9.81f, 0.0f));
	}

	return 1;
}

// Closes the physics system
int PhysicsClass::Close() {

	if(Enabled) {
		delete World;
		delete Solver;
		delete Dispatcher;
		delete BroadPhase;
		delete CollisionConfiguration;
	}

	return 1;
}

// Updates the physics system
void PhysicsClass::Update(u32 TDeltaTime) {

	if(Enabled) {

		World->stepSimulation(TDeltaTime * 0.001f * SimulationSpeed, 10);

		// Handle collision callbacks
		int ManifoldCount = World->getDispatcher()->getNumManifolds();
		for(int i = 0; i < ManifoldCount; i++) {
			btPersistentManifold *ContactManifold = World->getDispatcher()->getManifoldByIndexInternal(i);
			if(ContactManifold->getNumContacts() > 0) {
				btCollisionObject *CollisionObject0 = static_cast<btCollisionObject *>(ContactManifold->getBody0());
				btCollisionObject *CollisionObject1 = static_cast<btCollisionObject *>(ContactManifold->getBody1());
				
				ObjectClass *Object0 = static_cast<ObjectClass *>(CollisionObject0->getUserPointer());
				ObjectClass *Object1 = static_cast<ObjectClass *>(CollisionObject1->getUserPointer());

				Object0->HandleCollision(Object1, ContactManifold, 1);
				Object1->HandleCollision(Object0, ContactManifold, -1);
			}

		}
	}
}

// Performs raycasting on the world and returns the point of collision
bool PhysicsClass::RaycastWorld(const btVector3 &TStart, btVector3 &TEnd, btVector3 &TNormal) {

	if(Enabled) {
		btCollisionWorld::ClosestRayResultCallback RayCallback(TStart, TEnd);

		// Perform raycast
		World->rayTest(TStart, TEnd, RayCallback);
		if(RayCallback.hasHit()) {

			TEnd = RayCallback.m_hitPointWorld;
			TNormal = RayCallback.m_hitNormalWorld;
			return true;
		}
	}

	return false;
}

// Converts a position and euler rotation to matrix form
void PhysicsClass::GetBulletTransform(btTransform &TTransform, const vector3df &TPosition, const vector3df &TRotation) {

	// Rotation
	btQuaternion QuaternionRotation(TRotation.Y * DEGTORAD, TRotation.X * DEGTORAD, TRotation.Z * DEGTORAD);

	// Transform
	TTransform.setIdentity();
	TTransform.setOrigin(btVector3(TPosition.X, TPosition.Y, TPosition.Z));
	TTransform.setRotation(QuaternionRotation);
}

// Converts a quaternion to an euler angle
void PhysicsClass::QuaternionToEuler(const btQuaternion &TQuat, btVector3 &TEuler) {
	btScalar W = TQuat.getW();
	btScalar X = TQuat.getX();
	btScalar Y = TQuat.getY();
	btScalar Z = TQuat.getZ();
	float WSquared = W * W;
	float XSquared = X * X;
	float YSquared = Y * Y;
	float ZSquared = Z * Z;

	TEuler.setX(atan2f(2.0f * (Y * Z + X * W), -XSquared - YSquared + ZSquared + WSquared));
	TEuler.setY(asinf(-2.0f * (X * Z - Y * W)));
	TEuler.setZ(atan2f(2.0f * (X * Y + Z * W), XSquared - YSquared - ZSquared + WSquared));
	TEuler *= RADTODEG;
}

// Converts an irrlicht mesh to a bullet collision mesh
void PhysicsClass::ConvertCollisionMesh(IMesh *TMesh, btTriangleMesh *TCollisionMesh, const vector3df &TScale) {
	
	// Convert all mesh buffers
	for(u32 i = 0; i < TMesh->getMeshBufferCount(); i++)
		ConvertCollisionMeshBuffer(TMesh->getMeshBuffer(i), TCollisionMesh, TScale);
}

// Converts an irrlicht mesh buffer to a bullet collision mesh
void PhysicsClass::ConvertCollisionMeshBuffer(IMeshBuffer *TMeshBuffer, btTriangleMesh *TCollisionMesh, const vector3df &TScale) {
	
	btVector3 TriangleVertices[3];
	u16 *Indices = TMeshBuffer->getIndices();
	if(TMeshBuffer->getVertexType() == EVT_STANDARD) {
		
		S3DVertex *Vertices = (S3DVertex *)TMeshBuffer->getVertices();
		for(u32 i = 0; i < TMeshBuffer->getIndexCount(); i += 3) {
			TriangleVertices[0] = btVector3(Vertices[Indices[i]].Pos.X * TScale.X,
											Vertices[Indices[i]].Pos.Y * TScale.Y,
											Vertices[Indices[i]].Pos.Z * TScale.Z);
			TriangleVertices[1] = btVector3(Vertices[Indices[i+1]].Pos.X * TScale.X,
											Vertices[Indices[i+1]].Pos.Y * TScale.Y,
											Vertices[Indices[i+1]].Pos.Z * TScale.Z);
			TriangleVertices[2] = btVector3(Vertices[Indices[i+2]].Pos.X * TScale.X,
											Vertices[Indices[i+2]].Pos.Y * TScale.Y,
											Vertices[Indices[i+2]].Pos.Z * TScale.Z);

			TCollisionMesh->addTriangle(TriangleVertices[0], TriangleVertices[1], TriangleVertices[2]);
		}
	}
	else {
		
		S3DVertex2TCoords *Vertices = (S3DVertex2TCoords *)TMeshBuffer->getVertices();
		for(u32 i = 0; i < TMeshBuffer->getIndexCount(); i += 3) {
			TriangleVertices[0] = btVector3(Vertices[Indices[i]].Pos.X * TScale.X,
											Vertices[Indices[i]].Pos.Y * TScale.Y,
											Vertices[Indices[i]].Pos.Z * TScale.Z);
			TriangleVertices[1] = btVector3(Vertices[Indices[i+1]].Pos.X * TScale.X,
											Vertices[Indices[i+1]].Pos.Y * TScale.Y,
											Vertices[Indices[i+1]].Pos.Z * TScale.Z);
			TriangleVertices[2] = btVector3(Vertices[Indices[i+2]].Pos.X * TScale.X,
											Vertices[Indices[i+2]].Pos.Y * TScale.Y,
											Vertices[Indices[i+2]].Pos.Z * TScale.Z);

			TCollisionMesh->addTriangle(TriangleVertices[0], TriangleVertices[1], TriangleVertices[2]);
		}
	}
}

// Removes a bit field from a value
void PhysicsClass::RemoveFilter(int &TValue, int TFilter) {

	TValue &= (~TFilter);
}

// Sets default, static, or kinematic on a rigid body
void PhysicsClass::SetBodyType(int &TValue, int TFilter) {

	TValue &= (~FILTER_BASICBODIES);
	TValue |= TFilter;
}
