/*************************************************************************************
*	irrlamb - http://code.google.com/p/irrlamb/
*	Copyright (C) 2007  Alan Witkowski
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation, either version 3 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program.  If not, see <http://www.gnu.org/licenses/>.
**************************************************************************************/
#include "replay.h"
#include "save.h"
#include "globals.h"
#include "config.h"

using namespace std;

void ReplayClass::StartRecording(const stringc &TLevelFile, int TLevelVersion) {

	State = STATE_RECORDING;
	RecordInterval = Config::Instance().ReplayInterval;
	NextPacketTime = 0;
	Time = 0;

	// Create file
	LastReplayFile = Save::Instance().GetReplayPath() / "last.replay";
	ReplayOut.open(LastReplayFile.string().c_str(), ios::out | ios::binary);

	// Create header
	ReplayHeaderStruct Header;
	strcpy(Header.LevelFile, TLevelFile.c_str());
	strcpy(Header.Description, "No Description");
	strcpy(Header.Date, Save::Instance().GetDate().c_str());
	Header.ReplayVersion = REPLAY_VERSION;
	Header.LevelVersion = TLevelVersion;
	Header.CompletionTime = 0;
	Header.RecordInterval = RecordInterval;

	// Write header
	ReplayOut.write(reinterpret_cast<char *>(&Header), sizeof(Header));
}

// Stops the recording process
void ReplayClass::StopRecording() {
	
	State = STATE_NONE;
	ReplayOut.close();
}

// Makes a copy of last.replay and updates the header
bool ReplayClass::SaveReplay(const stringc &TDescription) {

	// Generate a new replay file
	stringc ReplayName = stringc(Save::Instance().GetNextReplayNumber()) + stringc(".replay");
	fs::path ReplayFile = Save::Instance().GetReplayPath() / ReplayName.c_str();

	// Flush the replay
	ReplayOut.flush();

	// Make a copy of the file
	if(fs::exists(ReplayFile))
		fs::remove(ReplayFile);
		
	fs::copy_file(LastReplayFile, ReplayFile);
	Save::Instance().IncrementReplayNumber();

	// Update the header
	ReplayHeaderStruct Header;
	if(!GetReplayHeader(ReplayFile.string().c_str(), Header))
		return false;

	strcpy(Header.Description, TDescription.c_str());
	Header.CompletionTime = Time;

	// Save new header
	if(!SaveReplayHeader(ReplayFile.string().c_str(), Header))
		return false;

	return true;
}

// Updates the replay timer
void ReplayClass::Update(u32 TDeltaTime) {

	Time += TDeltaTime;
	NextPacketTime += TDeltaTime;
}

// Determines if a packet is required
bool ReplayClass::NeedsPacket() {
	
	return State == STATE_RECORDING && NextPacketTime >= RecordInterval;
}

// Resets the next packet timer
void ReplayClass::ResetNextPacketTimer() {

	if(NeedsPacket())
		NextPacketTime = 0;
}

// Starts replay
bool ReplayClass::StartReplay(const stringc &TReplayFile, ReplayHeaderStruct &THeader) {

	// Open the replay
	ReplayIn.clear();
	ReplayIn.open((Save::Instance().GetReplayPath() / TReplayFile.c_str()).string().c_str(), ios::in | ios::binary);
	if(!ReplayIn.is_open())
		return false;

	// Read the header
	ReplayIn.read(reinterpret_cast<char *>(&THeader), sizeof(THeader));

	State = STATE_REPLAYING;

	return true;
}

// Stops replay
void ReplayClass::StopReplay() {
	
	State = STATE_NONE;
	ReplayIn.close();
}

// Returns true if the replay is done playing
bool ReplayClass::ReplayStopped() {

	return ReplayIn.eof();
}

// Retrieves the header of a replay
bool ReplayClass::GetReplayHeader(const stringc &TReplayFile, ReplayHeaderStruct &THeader) {
	
	// Open the replay
	IReadFile *ReadFile = irrFile->createAndOpenFile(TReplayFile.c_str());
	if(!ReadFile)
		return false;

	// Read the header
	ReadFile->read(&THeader, sizeof(THeader));
	ReadFile->drop();

	return true;
}

// Write the replay header
bool ReplayClass::SaveReplayHeader(const stringc &TReplayFile, ReplayHeaderStruct &THeader) {
	
	// Open the replay
	fstream FileOut(TReplayFile.c_str(), ios::in | ios::out | ios::binary);
	if(!FileOut)
		return false;

	// Write the header
	FileOut.seekp(0);
	FileOut.write(reinterpret_cast<char *>(&THeader), sizeof(THeader));
	FileOut.close();

	return true;
}

// Writes a packet header
void ReplayClass::WriteEvent(PacketType TType) {
	
	// Create the packet
	ReplayEventStruct Packet;
	Packet.Type = TType;
	Packet.TimeStamp = Time;
	
	// Write the packet
	ReplayOut.write(reinterpret_cast<char *>(&Packet), sizeof(Packet));

	//printf("Writing Packet Header: type=%d, time=%d\n", Packet.Type, Packet.TimeStamp);
}

// Reads a packet header
void ReplayClass::ReadEvent(ReplayEventStruct &TPacket) {

	ReplayIn.read(reinterpret_cast<char *>(&TPacket), sizeof(TPacket));
}
