/*************************************************************************************
*	irrlamb - http://code.google.com/p/irrlamb/
*	Copyright (C) 2007  Alan Witkowski
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation, either version 3 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program.  If not, see <http://www.gnu.org/licenses/>.
**************************************************************************************/
#ifndef REPLAY_H
#define REPLAY_H

// Libraries
#include "singleton.h"
#include <fstream>
#include <irrlicht/irrlicht.h>
#include <boost/filesystem/path.hpp>
#include <boost/filesystem/convenience.hpp>

const int REPLAY_VERSION = 2;

// Namespaces
using namespace irr;
using namespace core;
using namespace scene;
using namespace video;
using namespace io;
using namespace gui;
namespace fs = boost::filesystem;

// Structures
struct ReplayHeaderStruct {
	int ReplayVersion;
	int LevelVersion;
	char LevelFile[34];
	char Description[34];
	char Date[12];
	u32 CompletionTime;
	u32 RecordInterval;
};

// Event packet structure
struct ReplayEventStruct {
	u32 TimeStamp;
	int Type;
};

// This structure is used as a header for object data
struct ReplayObjectHeaderStruct {
	int Count;
};

// This contains movement data for each object
struct ReplayMovementStruct {
	int ObjectIndex;
	vector3df Position;
	vector3df Rotation;
};

// Object creation
struct ReplayObjectCreateStruct {
	int Template;
	vector3df Position;
	vector3df Rotation;
};

// Object deletion
struct ReplayObjectDeleteStruct {
	int ObjectID;
};

// Orb deactivation
struct ReplayOrbDeactivateStruct {
	int ObjectID;
	u32 Length;
};

// This contains camera data
struct ReplayCameraStruct {
	vector3df Position;
	vector3df LookAt;
};

// Classes
class ReplayClass {

	public:

		enum PacketType {
			PACKET_CAMERA,
			PACKET_MOVEMENT,
			PACKET_CREATE,
			PACKET_DELETE,
			PACKET_ORBDEACTIVATE,
		};

		enum StateType {
			STATE_NONE,
			STATE_RECORDING,
			STATE_REPLAYING,
		};

		void StartRecording(const stringc &TLevelFile, int TLevelVersion);
		void StopRecording();
		bool SaveReplay(const stringc &TDescription);

		bool StartReplay(const stringc &TReplayFile, ReplayHeaderStruct &THeader);
		void StopReplay();
		bool ReplayStopped();

		bool GetReplayHeader(const stringc &TReplayFile, ReplayHeaderStruct &THeader);
		bool SaveReplayHeader(const stringc &TReplayFile, ReplayHeaderStruct &THeader);

		void Update(u32 TDeltaTime);

		bool IsRecording() const { return State == STATE_RECORDING; }
		bool IsReplaying() const { return State == STATE_REPLAYING; }
		bool NeedsPacket();
		void ResetNextPacketTimer();

		void WriteEvent(PacketType TType);
		void ReadEvent(ReplayEventStruct &TPacket);
		template<typename T> void WritePacket(T &TPacket) {
			ReplayOut.write(reinterpret_cast<char *>(&TPacket), sizeof(TPacket));
		}
		template<typename T> void ReadPacket(T &TPacket) {
			ReplayIn.read(reinterpret_cast<char *>(&TPacket), sizeof(TPacket));
		}

	private:

		// Replays
		std::ofstream ReplayOut;
		std::ifstream ReplayIn;
		fs::path LastReplayFile;

		// Time management
		u32 Time, NextPacketTime, RecordInterval;

		// State
		StateType State;

};

// Singletons
typedef Singleton<ReplayClass> Replay;

#endif
