/*************************************************************************************
*	irrlamb - http://code.google.com/p/irrlamb/
*	Copyright (C) 2007  Alan Witkowski
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation, either version 3 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program.  If not, see <http://www.gnu.org/licenses/>.
**************************************************************************************/
#include "save.h"
#include "globals.h"
#include "../tinyxml/tinyxml.h"
#include <boost/date_time/gregorian/gregorian.hpp>

const int MAX_SCORES = 10;

using namespace boost::gregorian;

// Initializes the save system
int SaveClass::Init() {
	
	// Get save directories
	#ifdef _WIN32
		try {
			SavePath = fs::path(getenv("HOMEPATH")) / "My Documents";
			fs::create_directory(SavePath);
			SavePath /= "My Games";
			fs::create_directory(SavePath);
			SavePath /= "irrlamb";
		} catch(...) { }

	#else
		SavePath = fs::path(getenv("HOME")) / fs::path(".irrlamb", fs::native);
	#endif

	ReplayPath = SavePath / "replays";
	ScreenshotsPath = SavePath / "screenshots";

	// Get files
	ConfigFile = SavePath / "config.xml";
	StatsFile = SavePath / "stats.xml";

	// Create directories
	try {
		fs::create_directory(SavePath);
		fs::create_directory(ReplayPath);
		fs::create_directory(ScreenshotsPath);
	} catch(...) { }

	return 1;
}

// Closes the save system
int SaveClass::Close() {

	return 1;
}

// Finds the highest replay number
void SaveClass::FindHighestReplayNumber() {

	int HighCount = 0;

	// Change directory
	stringc OldWorkingDirectory = irrFile->getWorkingDirectory();
	irrFile->changeWorkingDirectoryTo(ReplayPath.string().c_str());
	
	// Get files
	IFileList *FileList = irrFile->createFileList();
	int FileCount = FileList->getFileCount();
	for(int i = 0; i < FileCount; i++) {
		if(!FileList->isDirectory(i)) {
			stringc File(FileList->getFileName(i));

			// Get the replay as a number
			int ExtensionIndex = File.find(".replay");
			if(ExtensionIndex != -1) {
				int Number = atoi(File.subString(0, ExtensionIndex).c_str());
				if(Number > HighCount)
					HighCount = Number;
			}
		}
	}
	irrFile->changeWorkingDirectoryTo(OldWorkingDirectory.c_str());

	// Get the highest replay number
	NextReplayNumber = HighCount + 1;
}

// Returns today's date
const stringc &SaveClass::GetDate() {
	static stringc Date;
	date Today = day_clock::local_day();
	Date = to_iso_extended_string(Today).c_str();
	
	return Date;
}

// Populates the level stats map
int SaveClass::LoadLevelStats() {

	// Open the XML file
	TiXmlDocument Document(StatsFile.string().c_str());
	if(!Document.LoadFile()) {
		printf("line %d column %d: %s\n", Document.ErrorRow(), Document.ErrorCol(), Document.ErrorDesc());
		return 0;
	}

	LevelStats.clear();

	// Check for stats tag
	TiXmlElement *StatsElement = Document.FirstChildElement("stats");
	if(!StatsElement)
		return 0;

	// Check for levels tag
	TiXmlElement *LevelsElement = StatsElement->FirstChildElement("levels");
	if(LevelsElement) {
		
		for(TiXmlElement *LevelElement = LevelsElement->FirstChildElement("level"); LevelElement != 0; LevelElement = LevelElement->NextSiblingElement("level")) {

			// Start level information
			stringc FileName = LevelElement->Attribute("file");
			SaveLevelStruct Stat;

			// Unlocks
			TiXmlElement *UnlockElement = LevelElement->FirstChildElement("unlock");
			UnlockElement->QueryIntAttribute("level", &Stat.UnlockLevel);

			// Counts
			TiXmlElement *CountElement = LevelElement->FirstChildElement("count");
			CountElement->QueryIntAttribute("load", &Stat.LoadCount);
			CountElement->QueryIntAttribute("lose", &Stat.LoseCount);
			CountElement->QueryIntAttribute("win", &Stat.WinCount);

			// Time
			TiXmlElement *TimeElement = LevelElement->FirstChildElement("time");
			TimeElement->QueryIntAttribute("play", (int *)(&Stat.PlayTime));

			// Scores
			TiXmlElement *ScoresElement = LevelElement->FirstChildElement("scores");
			if(ScoresElement) {
				int ScoreCount = 0;
				for(TiXmlElement *ScoreElement = ScoresElement->FirstChildElement("score"); ScoreElement != 0 && ScoreCount < MAX_SCORES; ScoreElement = ScoreElement->NextSiblingElement("score"), ScoreCount++) {
					int Time = atoi(ScoreElement->GetText());
					if(Time > 0)
						Stat.HighScores.push_back(SaveScoreStruct(Time, ScoreElement->Attribute("date")));
				}
			}

			// Add level information
			if(LevelStats.find(FileName) == LevelStats.end())
				LevelStats[FileName] = Stat;

		}

	}

	return 1;
}

// Save the map out to an xml file
void SaveClass::SaveLevelStats() {
	
	TiXmlDocument Document;

	// Create header
	Document.LinkEndChild(new TiXmlDeclaration("1.0", "", ""));
	
	// Create stats element
	TiXmlElement *StatsElement = new TiXmlElement("stats");
	StatsElement->SetAttribute("version", "1.0");
	Document.LinkEndChild(StatsElement);

	// Create video element
	TiXmlElement *LevelsElement = new TiXmlElement("levels");
	StatsElement->LinkEndChild(LevelsElement);

	SaveLevelStruct *Stat;
	for(LevelStatsIterator = LevelStats.begin(); LevelStatsIterator != LevelStats.end(); ++LevelStatsIterator) {
		Stat = &LevelStatsIterator->second;
		TiXmlElement *LevelElement = new TiXmlElement("level");
		LevelElement->SetAttribute("file", LevelStatsIterator->first.c_str());
		LevelsElement->LinkEndChild(LevelElement);

		// Unlocks
		TiXmlElement *UnlockElement = new TiXmlElement("unlock");
		UnlockElement->SetAttribute("level", Stat->UnlockLevel);
		LevelElement->LinkEndChild(UnlockElement);

		// Counts
		TiXmlElement *CountElement = new TiXmlElement("count");
		CountElement->SetAttribute("load", Stat->LoadCount);
		CountElement->SetAttribute("lose", Stat->LoseCount);
		CountElement->SetAttribute("win", Stat->WinCount);
		LevelElement->LinkEndChild(CountElement);

		// Time
		TiXmlElement *TimeElement = new TiXmlElement("time");
		TimeElement->SetAttribute("play", Stat->PlayTime);
		LevelElement->LinkEndChild(TimeElement);

		// Scores
		TiXmlElement *ScoresElement = new TiXmlElement("scores");
		LevelElement->LinkEndChild(ScoresElement);

		// Write scores
		for(u32 i = 0; i < Stat->HighScores.size(); i++) {
			TiXmlElement *ScoreElement = new TiXmlElement("score");
			ScoreElement->SetAttribute("date", Stat->HighScores[i].Date.c_str());
			ScoreElement->LinkEndChild(new TiXmlText(stringc(Stat->HighScores[i].Time).c_str()));
			ScoresElement->LinkEndChild(ScoreElement);
		}

	}

	// Write file
	Document.SaveFile(StatsFile.string().c_str());
}

// Updates a level stat
void SaveClass::UpdateLevelStats(const stringc &TLevel, const SaveLevelStruct &TStats) {

	LevelStats[TLevel] = TStats;
}

// Retrieves stats for a level
bool SaveClass::GetLevelStats(const stringc &TLevel, SaveLevelStruct &TStats) {

	if(LevelStats.find(TLevel) == LevelStats.end())
		return false;

	TStats = LevelStats[TLevel];

	return true;
}

// Adds a score to the list if it's high enough
void SaveClass::AddScore(const stringc &TLevel, u32 TTime) {

	SaveLevelStruct Stats;
	GetLevelStats(TLevel, Stats);

	// Add score
	Stats.HighScores.push_back(SaveScoreStruct(TTime, GetDate()));
	Stats.HighScores.sort();

	// Cut off old scores
	if(Stats.HighScores.size() > MAX_SCORES)
		Stats.HighScores.erase(MAX_SCORES, Stats.HighScores.size() - MAX_SCORES);

	UpdateLevelStats(TLevel, Stats);
}

// Increment the number of times a level has been loaded
void SaveClass::IncrementLevelLoadCount(const stringc &TLevel) {

	SaveLevelStruct Stats;
	GetLevelStats(TLevel, Stats);

	Stats.LoadCount++;
	UpdateLevelStats(TLevel, Stats);
}

// Increment the number of times the player has lost
void SaveClass::IncrementLevelLoseCount(const stringc &TLevel) {

	SaveLevelStruct Stats;
	GetLevelStats(TLevel, Stats);

	Stats.LoseCount++;
	UpdateLevelStats(TLevel, Stats);
}

// Increment the number of times the player has won
void SaveClass::IncrementLevelWinCount(const stringc &TLevel) {

	SaveLevelStruct Stats;
	GetLevelStats(TLevel, Stats);

	Stats.WinCount++;
	UpdateLevelStats(TLevel, Stats);
}

// Increment the total playing time for a level
void SaveClass::IncrementLevelPlayTime(const stringc &TLevel, u32 TTime) {

	SaveLevelStruct Stats;
	GetLevelStats(TLevel, Stats);

	Stats.PlayTime += TTime;
	UpdateLevelStats(TLevel, Stats);
}
