/*************************************************************************************
*	irrlamb - http://code.google.com/p/irrlamb/
*	Copyright (C) 2007  Alan Witkowski
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation, either version 3 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program.  If not, see <http://www.gnu.org/licenses/>.
**************************************************************************************/
#include "scripting.h"
#include "graphics.h"
#include "interface.h"
#include "random.h"
#include "objectmanager.h"
#include "game.h"
#include "level.h"
#include "replay.h"
#include "audio.h"
#include "save.h"
#include "campaign.h"
#include "../objects/default.h"
#include "../objects/player.h"
#include "../objects/orb.h"
#include "../objects/zone.h"
#include "../objects/basicjoint.h"
#include "../objects/springjoint.h"
#include "../play.h"

// Controls
luaL_Reg ScriptingClass::CameraFunctions[] = {
	{"SetYaw", &ScriptingClass::CameraSetYaw},
	{"SetPitch", &ScriptingClass::CameraSetPitch},
	{NULL, NULL}
};

// Functions for creating objects
luaL_Reg ScriptingClass::ObjectFunctions[] = {
	{"GetPointer", &ScriptingClass::ObjectGetPointer},
	{"GetName", &ScriptingClass::ObjectGetName},
	{"SetLifetime", &ScriptingClass::ObjectSetLifetime},
	{"Delete", &ScriptingClass::ObjectDelete},
	{NULL, NULL}
};

// Functions for manipulating rigid bodies
luaL_Reg ScriptingClass::RigidBodyFunctions[] = {
	{"Stop", &ScriptingClass::RigidBodyStop},
	{"SetPosition", &ScriptingClass::RigidBodySetPosition},
	{"GetPosition", &ScriptingClass::RigidBodyGetPosition},	
	{NULL, NULL}
};

// Functions for manipulating orbs
luaL_Reg ScriptingClass::OrbFunctions[] = {
	{"Deactivate", &ScriptingClass::OrbDeactivate},
	{NULL, NULL}
};


// Functions for timers
luaL_Reg ScriptingClass::TimerFunctions[] = {
	{"Stamp", &ScriptingClass::TimerStamp},
	{"DelayedFunction", &ScriptingClass::TimerDelayedFunction},
	{NULL, NULL}
};

// Functions for changing levels
luaL_Reg ScriptingClass::LevelFunctions[] = {
	{"Lose", &ScriptingClass::LevelLose},
	{"Win", &ScriptingClass::LevelWin},
	{"GetTemplate", &ScriptingClass::LevelGetTemplate},
	{"CreateObject", &ScriptingClass::LevelCreateObject},
	{"CreateJoint", &ScriptingClass::LevelCreateJoint},
	{"CreateSpring", &ScriptingClass::LevelCreateSpring},	
	{"SetPosition", &ScriptingClass::LevelSetPosition},
	{"SetRotation", &ScriptingClass::LevelSetRotation},
	{"UpdatePosition", &ScriptingClass::LevelUpdatePosition},
	{"UpdateRotation", &ScriptingClass::LevelUpdateRotation},
	{NULL, NULL}
};

// Functions for the GUI
luaL_Reg ScriptingClass::GUIFunctions[] = {
	{"TutorialText", &ScriptingClass::GUITutorialText},
	//{"UpdateText", &ScriptingClass::GUIUpdateText},
	//{"DeleteElement", &ScriptingClass::GUIDeleteElement},
	{NULL, NULL}
};

// Functions for random number generation
luaL_Reg ScriptingClass::RandomFunctions[] = {
	{"Seed", &ScriptingClass::RandomSeed},
	{"GetFloat", &ScriptingClass::RandomGetFloat},
	{"GetInt", &ScriptingClass::RandomGetInt},
	{NULL, NULL}
};

// Functions for manipulating zones
luaL_Reg ScriptingClass::ZoneFunctions[] = {
	
	{NULL, NULL}
};

/*
// Functions for audio
luaL_Reg ScriptingClass::AudioFunctions[] = {
	{"Play", &ScriptingClass::AudioPlay},
	{"Stop", &ScriptingClass::AudioStop},
	{NULL, NULL}
};
*/

// Constructor
ScriptingClass::ScriptingClass()
:	LuaObject(NULL) {
	
}

// Initializes the scripting interface
int ScriptingClass::Init() {

	// Restart scripting system
	Reset();

	return 1;
}

// Shuts down the scripting interface
int ScriptingClass::Close() {

	// Close old lua state
	if(LuaObject != NULL)
		lua_close(LuaObject);

	return 1;
}

// Resets the scripting state
void ScriptingClass::Reset() {
	
	// Close old lua state
	if(LuaObject != NULL)
		lua_close(LuaObject);
	
	// Initialize Lua object
	LuaObject = lua_open();
	luaopen_base(LuaObject);
	luaopen_math(LuaObject);

	// Register C++ functions used by Lua
	luaL_register(LuaObject, "Camera", CameraFunctions);
	luaL_register(LuaObject, "Object", ObjectFunctions);
	luaL_register(LuaObject, "RigidBody", RigidBodyFunctions);
	luaL_register(LuaObject, "Orb", OrbFunctions);
	luaL_register(LuaObject, "Timer", TimerFunctions);
	luaL_register(LuaObject, "Level", LevelFunctions);
	luaL_register(LuaObject, "GUI", GUIFunctions);
	luaL_register(LuaObject, "Random", RandomFunctions);
	luaL_register(LuaObject, "Zone", ZoneFunctions);

	// Clean up
	KeyCallbacks.clear();
	TimedCallbacks.clear();
}

// Loads a Lua file
int ScriptingClass::LoadFile(const stringc &TFile) {
	
	if(TFile == "")
		return 0;

	stringc Level = "scripts/";
	Level += TFile;
	
	// Load the file
	if(luaL_dofile(LuaObject, Level.c_str()) != 0) {
		stringc Buffer("Failed to load script: ");
		Buffer += lua_tostring(LuaObject, -1);
		irrLog->log(Buffer.c_str());
		return 0;
	}

	return 1;
}

// Checks for a valid argument count
bool ScriptingClass::CheckArguments(lua_State *TLuaObject, int TRequired) {
	int ArgumentCount = lua_gettop(TLuaObject);

	// Check for arguments
	if(ArgumentCount != TRequired) {

		lua_Debug Record;
		lua_getstack(TLuaObject, 0, &Record);
		lua_getinfo(TLuaObject, "nl", &Record);

		char Buffer[80];
		sprintf(Buffer, "Function %s requires %d arguments\n", Record.name, TRequired);
		irrLog->log(Buffer);
		return false;
	}

	return true;
}

// Gets function arguments from a Lua function, returns false if there were errors
int ScriptingClass::GetArguments(lua_State *TLuaObject, int TRequired, float *TData) {

	// Validate argument count
	if(!CheckArguments(TLuaObject, TRequired))
		return 0;

	// Get arguments
	for(int i = 0; i < TRequired; i++)
		TData[i] = (float)lua_tonumber(TLuaObject, i + 1);

	return 1;
}

// Calls a Lua function by name
void ScriptingClass::CallFunction(const stringc &TFunctionName) {
	
	// Check for the function name
	lua_getglobal(LuaObject, TFunctionName.c_str());
	if(!lua_isfunction(LuaObject, -1)) {
		lua_pop(LuaObject, 1);
		return;	
	}

	lua_call(LuaObject, 0, 0);
}

// Passes collision events to Lua
void ScriptingClass::CallCollisionHandler(const stringc &TFunctionName, ObjectClass *TBaseObject, ObjectClass *TOtherObject) {
	
	// Check for the function name
	lua_getglobal(LuaObject, TFunctionName.c_str());
	if(!lua_isfunction(LuaObject, -1)) {
		lua_pop(LuaObject, 1);
		return;	
	}
	
	lua_pushlightuserdata(LuaObject, static_cast<void *>(TBaseObject));
	lua_pushlightuserdata(LuaObject, static_cast<void *>(TOtherObject));
	lua_call(LuaObject, 2, 0);
}

// Calls a zone enter/exit event
void ScriptingClass::CallZoneHandler(const stringc &TFunctionName, int TType, ObjectClass *TZone, ObjectClass *TObject) {
	
	// Check for the function name
	lua_getglobal(LuaObject, TFunctionName.c_str());
	if(!lua_isfunction(LuaObject, -1)) {
		lua_pop(LuaObject, 1);
		return;	
	}
	
	lua_pushinteger(LuaObject, TType);
	lua_pushlightuserdata(LuaObject, static_cast<void *>(TZone));
	lua_pushlightuserdata(LuaObject, static_cast<void *>(TObject));
	lua_call(LuaObject, 3, 1);

	int Disable = (int)lua_tointeger(LuaObject, -1);
	if(Disable) {
		ZoneClass *Zone = static_cast<ZoneClass *>(TZone);
		Zone->SetActive(false);
	}
}

// Calls a Lua function from a given keyname
bool ScriptingClass::HandleKeyPress(EKEY_CODE TKey) {

	// Find function
	KeyCallbacksIterator = KeyCallbacks.find(TKey);
	if(KeyCallbacksIterator != KeyCallbacks.end()) {
		CallFunction(KeyCallbacksIterator->second);
		return true;
	}

	return false;
}

// Passes mouse clicks to Lua
void ScriptingClass::HandleMousePress(int TButton, int TMouseX, int TMouseY) {

	// Get Lua function
	lua_getglobal(LuaObject, "HandleMousePress");
	if(!lua_isfunction(LuaObject, -1)) {
		lua_pop(LuaObject, 1);
		return;
	}

	// Pass parameters and call function
	lua_pushnumber(LuaObject, TButton);
	lua_pushnumber(LuaObject, TMouseX);
	lua_pushnumber(LuaObject, TMouseY);
	lua_call(LuaObject, 3, 0);		
}

// Sets the camera's yaw value
int ScriptingClass::CameraSetYaw(lua_State *TLuaObject) {
	
	// Validate arguments
	if(!CheckArguments(TLuaObject, 1))
		return 0;
	
	float Yaw = (float)lua_tonumber(TLuaObject, 1);

	// TODO: clean up
	if(PlayState::Instance()->GetCamera())
		PlayState::Instance()->GetCamera()->SetYaw(Yaw);

	return 0;
}

// Sets the camera's yaw value
int ScriptingClass::CameraSetPitch(lua_State *TLuaObject) {
	
	// Validate arguments
	if(!CheckArguments(TLuaObject, 1))
		return 0;
	
	float Pitch = (float)lua_tonumber(TLuaObject, 1);

	if(PlayState::Instance()->GetCamera())
		PlayState::Instance()->GetCamera()->SetPitch(Pitch);

	return 0;
}

// Gets a pointer to an object from a name
int ScriptingClass::ObjectGetPointer(lua_State *TLuaObject) {
	
	// Validate arguments
	if(!CheckArguments(TLuaObject, 1))
		return 0;

	stringc Name = lua_tostring(TLuaObject, 1);
	ObjectClass *Object = ObjectManager::Instance().GetObjectByName(Name);

	// Pass pointer
	lua_pushlightuserdata(TLuaObject, (void *)Object);

	return 1;
}

// Gets the name of an object
int ScriptingClass::ObjectGetName(lua_State *TLuaObject) {
	
	// Validate arguments
	if(!CheckArguments(TLuaObject, 1))
		return 0;

	ObjectClass *Object = (ObjectClass *)(lua_touserdata(TLuaObject, 1));
	
	if(Object != NULL)
		lua_pushstring(TLuaObject, Object->GetName().c_str());

	return 1;
}

// Sets the object's lifetime
int ScriptingClass::ObjectSetLifetime(lua_State *TLuaObject) {
	
	// Validate arguments
	if(!CheckArguments(TLuaObject, 2))
		return 0;

	ObjectClass *Object = (ObjectClass *)(lua_touserdata(TLuaObject, 1));
	float Lifetime = (float)lua_tonumber(TLuaObject, 2);
	if(Object != NULL)
		Object->SetLifetime(Lifetime);

	return 0;
}

// Sets the rigid body's position
int ScriptingClass::RigidBodySetPosition(lua_State *TLuaObject) {
	
	if(Replay::Instance().IsReplaying())
		return 0;

	// Validate arguments
	if(!CheckArguments(TLuaObject, 4))
		return 0;

	RigidBodyClass *RigidBody = (RigidBodyClass *)(lua_touserdata(TLuaObject, 1));
	float PositionX = (float)lua_tonumber(TLuaObject, 2);
	float PositionY = (float)lua_tonumber(TLuaObject, 3);
	float PositionZ = (float)lua_tonumber(TLuaObject, 4);
	if(RigidBody != NULL)
		RigidBody->SetPosition(btVector3(PositionX, PositionY, PositionZ));

	return 0;
}

// Gets the rigid body's position
int ScriptingClass::RigidBodyGetPosition(lua_State *TLuaObject) {
	
	if(Replay::Instance().IsReplaying()) {
		lua_pushnumber(TLuaObject, 0);
		lua_pushnumber(TLuaObject, 1);
		lua_pushnumber(TLuaObject, 2);
		return 3;
	}

	// Validate arguments
	if(!CheckArguments(TLuaObject, 1))
		return 0;

	RigidBodyClass *RigidBody = (RigidBodyClass *)(lua_touserdata(TLuaObject, 1));
	if(RigidBody == NULL)
		return 0;

	lua_pushnumber(TLuaObject, RigidBody->GetPosition()[0]);
	lua_pushnumber(TLuaObject, RigidBody->GetPosition()[1]);
	lua_pushnumber(TLuaObject, RigidBody->GetPosition()[2]);

	return 3;
}

// Deletes an object
int ScriptingClass::ObjectDelete(lua_State *TLuaObject) {
	
	// Validate arguments
	if(!CheckArguments(TLuaObject, 1))
		return 0;

	ObjectClass *Object = (ObjectClass *)(lua_touserdata(TLuaObject, 1));
	if(Object != NULL)
		ObjectManager::Instance().DeleteObject(Object);

	return 0;
}

// Stops a rigid body's movement
int ScriptingClass::RigidBodyStop(lua_State *TLuaObject) {
	
	// Validate arguments
	if(!CheckArguments(TLuaObject, 1))
		return 0;

	RigidBodyClass *RigidBody = (RigidBodyClass *)(lua_touserdata(TLuaObject, 1));
	if(RigidBody != NULL)
		RigidBody->Stop();

	return 0;
}

// Deactivates an orb
int ScriptingClass::OrbDeactivate(lua_State *TLuaObject) {
	
	// Validate arguments
	if(!CheckArguments(TLuaObject, 3))
		return 0;

	OrbClass *Orb = (OrbClass *)(lua_touserdata(TLuaObject, 1));
	stringc FunctionName = lua_tostring(TLuaObject, 2);
	float Time = (float)lua_tonumber(TLuaObject, 3);

	if(Orb != NULL && !Orb->GetDeactivated() && !Orb->GetDeactivating()) {
		Orb->StartDeactivation(FunctionName, (u32)(Time * 1000.0f));
		Audio::Instance().Play("game_deactivate");
	}

	return 0;
}

// Returns a time stamp from the start of the level
int ScriptingClass::TimerStamp(lua_State *TLuaObject) {

	lua_pushnumber(TLuaObject, PlayState::Instance()->GetTimer().GetSeconds());

	return 1;
}

// Adds a timed callback
int ScriptingClass::TimerDelayedFunction(lua_State *TLuaObject) {

	// Validate arguments
	if(!CheckArguments(TLuaObject, 2))
		return 0;

	stringc FunctionName = lua_tostring(TLuaObject, 1);
	float Time = (float)lua_tonumber(TLuaObject, 2);

	Scripting::Instance().AddTimedCallback(FunctionName, Time);

	return 0;
}

// Restarts the level
int ScriptingClass::LevelLose(lua_State *TLuaObject) {

	Save::Instance().IncrementLevelLoseCount(PlayState::Instance()->GetCurrentLevel());
	PlayState::Instance()->InitLose();
	
	return 0;
}

// Wins the level
int ScriptingClass::LevelWin(lua_State *TLuaObject) {

	// Increment win count
	SaveLevelStruct CurrentStats;
	Save::Instance().GetLevelStats(PlayState::Instance()->GetCurrentLevel(), CurrentStats);
	CurrentStats.WinCount++;
	Save::Instance().UpdateLevelStats(PlayState::Instance()->GetCurrentLevel(), CurrentStats);

	// Add high score
	Save::Instance().AddScore(PlayState::Instance()->GetCurrentLevel(), PlayState::Instance()->GetTimer().GetTime());

	// Get next level
	if(Campaign::Instance().IncrementLevel() && !Game::Instance().GetQuitOnEscape()) {

		// Unlock next level
		SaveLevelStruct Stats;
		Save::Instance().GetLevelStats(Campaign::Instance().GetLevel(), Stats);
		if(Stats.UnlockLevel < CurrentStats.UnlockLevel)
			Stats.UnlockLevel++;
		Save::Instance().UpdateLevelStats(Campaign::Instance().GetLevel(), Stats);

		PlayState::Instance()->SetNextLevel(Campaign::Instance().GetLevel());
	}
	else
		PlayState::Instance()->SetNextLevel("");

	PlayState::Instance()->InitWin();
	
	return 0;
}

// Gets a template from a name
int ScriptingClass::LevelGetTemplate(lua_State *TLuaObject) {

	// Validate arguments
	if(!CheckArguments(TLuaObject, 1))
		return 0;

	stringc TemplateName = lua_tostring(TLuaObject, 1);

	lua_pushinteger(TLuaObject, Level::Instance().GetTemplateFromName(TemplateName));
	
	return 1;
}

// Creates an object from a template
int ScriptingClass::LevelCreateObject(lua_State *TLuaObject) {

	// Validate arguments
	if(!CheckArguments(TLuaObject, 1))
		return 0;

	int TemplateIndex = (int)lua_tointeger(TLuaObject, 1);

	ObjectClass *Object = Level::Instance().CreateObjectFromTemplateIndex(TemplateIndex);
	lua_pushlightuserdata(TLuaObject, static_cast<void *>(Object));

	return 1;
}

// Creates a constraint
int ScriptingClass::LevelCreateJoint(lua_State *TLuaObject) {

	// Validate arguments
	if(!CheckArguments(TLuaObject, 3))
		return 0;

	int TemplateIndex = (int)lua_tointeger(TLuaObject, 1);

	DefaultObjectStruct ObjectData;
	if(!Level::Instance().GetTemplateFromIndex(TemplateIndex, ObjectData) || ObjectData.Type != ObjectClass::BASICJOINT)
		return 0;

	ObjectData.BodyA = (RigidBodyClass *)(lua_touserdata(TLuaObject, 2));
	ObjectData.BodyB = (RigidBodyClass *)(lua_touserdata(TLuaObject, 3));

	Level::Instance().WriteCreateEvent(TemplateIndex);

	ObjectClass *Object = ObjectManager::Instance().AddObject(new BasicJointClass(ObjectData));

	lua_pushlightuserdata(TLuaObject, static_cast<void *>(Object));

	return 1;
}

// Creates a spring constraint
int ScriptingClass::LevelCreateSpring(lua_State *TLuaObject) {

	// Validate arguments
	if(!CheckArguments(TLuaObject, 3))
		return 0;

	int TemplateIndex = (int)lua_tointeger(TLuaObject, 1);

	DefaultObjectStruct ObjectData;
	if(!Level::Instance().GetTemplateFromIndex(TemplateIndex, ObjectData) || ObjectData.Type != ObjectClass::SPRINGJOINT)
		return 0;

	ObjectData.BodyA = (RigidBodyClass *)(lua_touserdata(TLuaObject, 2));
	ObjectData.BodyB = (RigidBodyClass *)(lua_touserdata(TLuaObject, 3));

	Level::Instance().WriteCreateEvent(TemplateIndex);

	ObjectClass *Object = ObjectManager::Instance().AddObject(new SpringJointClass(ObjectData));

	lua_pushlightuserdata(TLuaObject, static_cast<void *>(Object));

	return 1;
}

// Sets the default object position
int ScriptingClass::LevelSetPosition(lua_State *TLuaObject) {

	// Check for arguments
	float Data[3];
	if(!GetArguments(TLuaObject, 3, Data))
		return 0;

	// Set value
	Level::Instance().SetDefaultPosition(vector3df(Data[0], Data[1], Data[2]));

	return 0;
}

// Sets the default object rotation
int ScriptingClass::LevelSetRotation(lua_State *TLuaObject) {

	// Check for arguments
	float Data[3];
	if(!GetArguments(TLuaObject, 3, Data))
		return 0;

	// Set value
	Level::Instance().SetDefaultRotation(vector3df(Data[0], Data[1], Data[2]));

	return 0;
}

// Updates the default object position
int ScriptingClass::LevelUpdatePosition(lua_State *TLuaObject) {

	// Check for arguments
	float Data[3];
	if(!GetArguments(TLuaObject, 3, Data))
		return 0;

	// Set value
	Level::Instance().UpdateDefaultPosition(vector3df(Data[0], Data[1], Data[2]));

	return 0;
}

// Updates the default object rotation
int ScriptingClass::LevelUpdateRotation(lua_State *TLuaObject) {

	// Check for arguments
	float Data[3];
	if(!GetArguments(TLuaObject, 3, Data))
		return 0;

	// Set value
	Level::Instance().UpdateDefaultRotation(vector3df(Data[0], Data[1], Data[2]));

	return 0;
}

// Sets the tutorial text
int ScriptingClass::GUITutorialText(lua_State *TLuaObject) {

	// Validate arguments
	if(!CheckArguments(TLuaObject, 2))
		return 0;

	stringw Text(lua_tostring(TLuaObject, 1));
	float Length = (float)lua_tonumber(TLuaObject, 2);

	Interface::Instance().SetTutorialText(Text, Length);

	return 0;
}

// Sets the random seed
int ScriptingClass::RandomSeed(lua_State *TLuaObject) {

	// Validate arguments
	if(!CheckArguments(TLuaObject, 1))
		return 0;

	u32 Seed = (u32)lua_tointeger(TLuaObject, 1);

	Random::Instance().SetSeed(Seed);

	return 0;
}

// Generates a random float
int ScriptingClass::RandomGetFloat(lua_State *TLuaObject) {

	// Validate arguments
	if(!CheckArguments(TLuaObject, 2))
		return 0;

	float Min = (float)lua_tonumber(TLuaObject, 1);
	float Max = (float)lua_tonumber(TLuaObject, 2);

	lua_pushnumber(TLuaObject, Random::Instance().GenerateRange(Min, Max));

	return 1;
}

// Generates a random integer
int ScriptingClass::RandomGetInt(lua_State *TLuaObject) {

	// Validate arguments
	if(!CheckArguments(TLuaObject, 2))
		return 0;

	int Min = (int)lua_tointeger(TLuaObject, 1);
	int Max = (int)lua_tointeger(TLuaObject, 2);

	lua_pushnumber(TLuaObject, Random::Instance().GenerateRange(Min, Max));

	return 1;
}

// Adds a timed callback function to the list
void ScriptingClass::AddTimedCallback(const stringc &TFunctionName, float TTime) {

	// Check time
	if(TTime <= 0.0)
		return;

	// Create callback structure
	TimedCallbackStruct Callback;
	Callback.TimeStamp = PlayState::Instance()->GetTimer().GetSeconds() + TTime;
	Callback.Function = TFunctionName;
	
	// Insert in order
	list<TimedCallbackStruct>::Iterator Iterator;
	for(Iterator = TimedCallbacks.begin(); Iterator != TimedCallbacks.end(); ++Iterator) {
		if(Callback.TimeStamp < (*Iterator).TimeStamp)
			break;
	}

	// Hack for irrlicht bug
	if(Iterator == TimedCallbacks.end())
		TimedCallbacks.push_back(Callback);
	else
		TimedCallbacks.insert_before(Iterator, Callback);
}

// Updates the timed callback functions
void ScriptingClass::UpdateTimedCallbacks() {

	for(list<TimedCallbackStruct>::Iterator Iterator(TimedCallbacks.begin()); Iterator != TimedCallbacks.end(); ++Iterator) {
		if(PlayState::Instance()->GetTimer().GetSeconds() >= (*Iterator).TimeStamp) {
			Scripting::Instance().CallFunction((*Iterator).Function);

			// Remove callback
			Iterator = TimedCallbacks.erase(Iterator);
			if(Iterator == TimedCallbacks.end())
				break;
		}
		else
			return;
	}
}

// Attaches a key to a Lua function
void ScriptingClass::AttachKeyToFunction(EKEY_CODE TKey, const stringc &TFunctionName) {

	// Install callback
	KeyCallbacksIterator = KeyCallbacks.find(TKey);
	if(KeyCallbacksIterator == KeyCallbacks.end())
		KeyCallbacks.insert(std::pair<EKEY_CODE, stringc>(TKey, TFunctionName));

}
