/*************************************************************************************
*	irrlamb - http://code.google.com/p/irrlamb/
*	Copyright (C) 2007  Alan Witkowski
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation, either version 3 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program.  If not, see <http://www.gnu.org/licenses/>.
**************************************************************************************/
#include "menu.h"
#include "engine/graphics.h"
#include "engine/input.h"
#include "engine/interface.h"
#include "engine/globals.h"
#include "engine/scripting.h"
#include "engine/config.h"
#include "engine/game.h"
#include "engine/save.h"
#include "engine/campaign.h"
#include "engine/audio.h"
#include "viewreplay.h"
#include "play.h"
#include <boost/filesystem/path.hpp>
#include <boost/filesystem/convenience.hpp>

namespace fs = boost::filesystem;

const wchar_t *GameVersion = L"v0.0.5";
const int CAMPAIGN_LEVELID = 1000;
const int PLAY_CAMPAIGNID = 900;

// List of action names
const wchar_t *MenuState::ActionNames[ACTIONS::COUNT] = {
	L"Move Forward",
	L"Move Back",
	L"Move Left",
	L"Move Right",
	L"Reset",
	L"Jump"
};

// Initializes the state
int MenuState::Init() {

	State = STATE_INITMAIN;
	
	CampaignIndex = 0;
	SelectedLevel = -1;
	DoubleClickTimer = 0;

	Input::Instance().SetMouseLocked(false);
	irrDevice->getCursorControl()->setVisible(true);

	ReloadAudio();

	return 1;
}

// Shuts the state down
int MenuState::Close() {

	return 1;
}

// Key presses
bool MenuState::HandleKeyPress(EKEY_CODE TKey) {

	bool Processed = true;

	switch(State) {
		case STATE_MAIN:
			switch(TKey) {
				case KEY_ESCAPE:
					Game::Instance().SetDone(true);
				break;
				default:
					Processed = false;
				break;
			}
		break;
		case STATE_PLAY:
			switch(TKey) {
				case KEY_ESCAPE:
					State = STATE_INITMAIN;
				break;
				default:
					Processed = false;
				break;
			}
		break;
		case STATE_LEVELS:
			switch(TKey) {
				case KEY_ESCAPE:
					State = STATE_INITPLAY;
				break;
				default:
					Processed = false;
				break;
			}
		break;
		case STATE_REPLAYS:
			switch(TKey) {
				case KEY_ESCAPE:
					State = STATE_INITMAIN;
				break;
				default:
					Processed = false;
				break;
			}
		break;
		case STATE_OPTIONS:
			switch(TKey) {
				case KEY_ESCAPE:
					State = STATE_INITMAIN;
				break;
				default:
					Processed = false;
				break;
			}
		break;
		case STATE_VIDEO:
			switch(TKey) {
				case KEY_ESCAPE:
					State = STATE_INITOPTIONS;
				break;
				default:
					Processed = false;
				break;
			}
		break;
		case STATE_AUDIO:
			switch(TKey) {
				case KEY_ESCAPE:
					State = STATE_INITOPTIONS;
				break;
				default:
					Processed = false;
				break;
			}
		break;
		case STATE_CONTROLS:
			if(KeyButton != NULL) {
				stringw KeyName = Input::Instance().GetKeyName(TKey);

				// Assign the key
				if(TKey != KEY_ESCAPE && KeyName != "") {
					
					int ActionType = KeyButton->getID() - CONTROLS_MOVEFORWARD;

					// Swap if the key already exists
					for(int i = 0; i < ACTIONS::COUNT; i++) {
						if(CurrentKeys[i] == TKey) {
							
							// Get button
							IGUIButton *SwapButton = static_cast<IGUIButton *>(irrGUI->getRootGUIElement()->getElementFromId(CONTROLS_MOVEFORWARD + i));

							// Swap text
							SwapButton->setText(stringw(Input::Instance().GetKeyName(CurrentKeys[ActionType])).c_str());
							CurrentKeys[i] = CurrentKeys[ActionType];
							break;
						}
					}

					// Update key
					KeyButton->setText(KeyName.c_str());
					CurrentKeys[ActionType] = TKey;
				}
				else
					KeyButton->setText(KeyButtonOldText.c_str());

				KeyButton = NULL;
			}
			else {
				switch(TKey) {
					case KEY_ESCAPE:
						State = STATE_INITOPTIONS;
					break;
					default:
						Processed = false;
					break;
				}
			}
		break;
	}

	return Processed;
}

// Handles the mouse
bool MenuState::HandleMousePress(int TButton, int TMouseX, int TMouseY) {

	switch(State) {
		case STATE_LEVELS: {
			IGUIElement *Element = irrGUI->getRootGUIElement()->getElementFromPoint(position2di(TMouseX, TMouseY));
			if(Element && Element->getID() == LEVELS_SELECTEDLEVEL) {
				if(DoubleClickTimer < 300)
					LaunchLevel();

				DoubleClickTimer = 0;				
				return true;
			}
		}
		break;
	}

	return false;	
}

// Handles GUI events
void MenuState::HandleGUI(EGUI_EVENT_TYPE TEventType, IGUIElement *TElement) {

	switch(TEventType) {
		case EGET_BUTTON_CLICKED:
			switch(TElement->getID()) {
				case MAIN_TUTORIAL:
				break;
				case MAIN_PLAY:
					Audio::Instance().Play("menu_click");
					State = STATE_INITPLAY;
				break;
				case MAIN_REPLAYS:
					Audio::Instance().Play("menu_click");
					State = STATE_INITREPLAYS;
				break;
				case MAIN_OPTIONS:
					Audio::Instance().Play("menu_click");
					State = STATE_INITOPTIONS;
				break;
				case MAIN_QUIT:
					Game::Instance().SetDone(true);
				break;
				case PLAY_BACK:
					Audio::Instance().Play("menu_click");
					State = STATE_INITMAIN;
				break;
				case LEVELS_GO:
					Audio::Instance().Play("menu_click");
					LaunchLevel();
				break;
				case LEVELS_BACK:
					Audio::Instance().Play("menu_click");
					State = STATE_INITPLAY;
				break;
				case REPLAYS_GO:
					Audio::Instance().Play("menu_click");
					LaunchReplay();
				break;
				case REPLAYS_DELETE: {
					// Get list
					IGUIListBox *ReplayList = static_cast<IGUIListBox *>(irrGUI->getRootGUIElement()->getElementFromId(REPLAYS_FILES));
					int SelectedIndex = ReplayList->getSelected();
					if(SelectedIndex != -1) {

						// Remove file
						fs::path File = Save::Instance().GetReplayPath() / GetReplayFile().c_str();
						try {
							fs::remove(File);
						} catch(...) { }

						// Remove entry
						ReplayList->removeItem(SelectedIndex);
					}
				}
				break;
				case REPLAYS_BACK:
					Audio::Instance().Play("menu_click");
					State = STATE_INITMAIN;
				break;
				case OPTIONS_VIDEO:
					Audio::Instance().Play("menu_click");
					State = STATE_INITVIDEO;
				break;
				case OPTIONS_AUDIO:
					Audio::Instance().Play("menu_click");
					State = STATE_INITAUDIO;
				break;
				case OPTIONS_CONTROLS:
					Audio::Instance().Play("menu_click");
					State = STATE_INITCONTROLS;
				break;
				case OPTIONS_BACK:
					Audio::Instance().Play("menu_click");
					State = STATE_INITMAIN;
				break;
				case VIDEO_SAVE: {
					Audio::Instance().Play("menu_click");

					// Save the video mode
					IGUIComboBox *VideoModes = static_cast<IGUIComboBox *>(irrGUI->getRootGUIElement()->getElementFromId(VIDEO_VIDEOMODES));
					if(VideoModes != NULL) {
						VideoModeStruct Mode = Graphics::Instance().GetVideoModes()[VideoModes->getSelected()];
						Config::Instance().ScreenWidth = Mode.Width;
						Config::Instance().ScreenHeight = Mode.Height;
					}

					// Save full screen
					IGUICheckBox *FullScreen = static_cast<IGUICheckBox *>(irrGUI->getRootGUIElement()->getElementFromId(VIDEO_FULLSCREEN));
					Config::Instance().FullScreen = FullScreen->isChecked();

					// Write config
					Config::Instance().WriteConfig();

					State = STATE_INITOPTIONS;
				}
				break;
				case VIDEO_CANCEL:
					Audio::Instance().Play("menu_click");
					State = STATE_INITOPTIONS;
				break;
				case AUDIO_SAVE: {
					Audio::Instance().Play("menu_click");

					int OldAudioEnabled = Config::Instance().AudioEnabled;

					// Get settings
					IGUICheckBox *AudioEnabled = static_cast<IGUICheckBox *>(irrGUI->getRootGUIElement()->getElementFromId(AUDIO_ENABLED));
					Config::Instance().AudioEnabled = AudioEnabled->isChecked();

					IGUIScrollBar *SoundVolume = static_cast<IGUIScrollBar *>(irrGUI->getRootGUIElement()->getElementFromId(AUDIO_SOUND));
					Config::Instance().SoundVolume = SoundVolume->getPos() / 100.0f;

					IGUIScrollBar *MusicVolume = static_cast<IGUIScrollBar *>(irrGUI->getRootGUIElement()->getElementFromId(AUDIO_MUSIC));
					Config::Instance().MusicVolume = MusicVolume->getPos() / 100.0f;

					// Reset audio
					Audio::Instance().Close();
					Audio::Instance().Init(Config::Instance().AudioEnabled);
					ReloadAudio();

					Config::Instance().WriteConfig();

					State = STATE_INITOPTIONS;
				}
				break;
				case AUDIO_CANCEL:
					Audio::Instance().Play("menu_click");
					State = STATE_INITOPTIONS;
				break;
				case CONTROLS_SAVE: {
					Audio::Instance().Play("menu_click");

					// Write config
					for(int i = 0; i < ACTIONS::COUNT; i++)
						Config::Instance().Keys[i] = CurrentKeys[i];

					Config::Instance().WriteConfig();

					State = STATE_INITOPTIONS;
				}
				break;
				case CONTROLS_CANCEL:
					Audio::Instance().Play("menu_click");
					State = STATE_INITOPTIONS;
				break;
				case CONTROLS_MOVEFORWARD:
				case CONTROLS_MOVEBACK:
				case CONTROLS_MOVELEFT:
				case CONTROLS_MOVERIGHT:
				case CONTROLS_MOVERESET:
				case CONTROLS_MOVEJUMP:
				{
					Audio::Instance().Play("menu_click");
					if(KeyButton)
						CancelKeyBind();

					KeyButton = static_cast<IGUIButton *>(TElement);
					KeyButtonOldText = KeyButton->getText();
					KeyButton->setText(L"");
				}
				break;
				default: {

					if(TElement->getID() >= CAMPAIGN_LEVELID) {
						int ClickedIndex = TElement->getID() - CAMPAIGN_LEVELID;
						if(SelectedLevel != ClickedIndex) {
							SelectedLevel = ClickedIndex;
							UpdateSelectedLevel(SelectedLevel);
							UpdateLevelInfo(SelectedLevel);
							DoubleClickTimer = 0;
						}
					}
					else if(TElement->getID() >= PLAY_CAMPAIGNID) {
						CampaignIndex = TElement->getID() - PLAY_CAMPAIGNID;
						State = STATE_INITLEVELS;
					}
				}
				break;
			}
		break;
		case EGET_LISTBOX_SELECTED_AGAIN:
			switch(TElement->getID()) {
				case REPLAYS_FILES:
					LaunchReplay();
				break;
			}
		break;
	}
}

// Updates the current state
void MenuState::Update(u32 TDeltaTime) {
	DoubleClickTimer += TDeltaTime;

	switch(State) {
		case STATE_INITMAIN: {	
			Interface::Instance().Clear();
			//IGUIButton *ButtonTutorial = irrGUI->addButton(Interface::Instance().GetAbsoluteRect(0.4f, 0.3f, 0.6f, 0.35f), 0, MAIN_TUTORIAL, L"Tutorial");
			IGUIButton *ButtonNewGame = irrGUI->addButton(Interface::Instance().GetAbsoluteRectCentered(0.5f, 0.35f, 0.2f, 0.05f), 0, MAIN_PLAY, L"Play");
			IGUIButton *ButtonReplays = irrGUI->addButton(Interface::Instance().GetAbsoluteRectCentered(0.5f, 0.45f, 0.2f, 0.05f), 0, MAIN_REPLAYS, L"Replays");
			IGUIButton *ButtonOptions = irrGUI->addButton(Interface::Instance().GetAbsoluteRectCentered(0.5f, 0.55f, 0.2f, 0.05f), 0, MAIN_OPTIONS, L"Options");
			IGUIButton *ButtonQuit = irrGUI->addButton(Interface::Instance().GetAbsoluteRectCentered(0.5f, 0.65f, 0.2f, 0.05f), 0, MAIN_QUIT, L"Quit");
			irrGUI->addImage(irrDriver->getTexture("art/title_logo.jpg"), position2di(Interface::Instance().GetAbsoluteX(0.5f - Interface::Instance().GetFractionX(256)), Interface::Instance().GetAbsoluteY(0.05f)));

			IGUIStaticText *TextVersion = irrGUI->addStaticText(GameVersion, Interface::Instance().GetAbsoluteRectWH(0.02f, 0.95f, 0.1f, 0.05f), false, false);

			State = STATE_MAIN;
		}
		break;
		case STATE_MAIN:
		break;
		case STATE_INITPLAY: {
			Interface::Instance().Clear();

			const array<CampaignStruct> &Campaigns = Campaign::Instance().GetCampaigns();
			float Position = 0.3f;
			for(u32 i = 0; i < Campaigns.size(); i++) {
				irrGUI->addButton(Interface::Instance().GetAbsoluteRectCentered(0.5f, Position, 0.15f, 0.05f), 0, PLAY_CAMPAIGNID + i, Campaigns[i].Name.c_str());
				Position += 0.1f;
			}

			Position += 0.1f;
			irrGUI->addButton(Interface::Instance().GetAbsoluteRectCentered(0.5f, Position, 0.15f, 0.05f), 0, PLAY_BACK, L"Back");

			State = STATE_PLAY;
		}
		break;
		case STATE_INITLEVELS: {
			Interface::Instance().Clear();

			// Add level info box
			stringw LevelText = L"Level Locked";
			IGUIStaticText *TextInfo = irrGUI->addStaticText(LevelText.c_str(), Interface::Instance().GetAbsoluteRectWH(0.05f, 0.1f, 0.2f, 0.7f), true, true, 0, LEVELINFO_DESCRIPTION, true);

			float TextPosition[2] = { 0.15f, 0.2f };
			irrGUI->addStaticText(L"Attempts", Interface::Instance().GetRightTextRect(L"Attempts", TextPosition[0], TextPosition[1], TextPosition[1] + 0.03f), false, false, 0, -1);
			irrGUI->addStaticText(L"0", Interface::Instance().GetAbsoluteRectWH(TextPosition[0] + 0.02f, TextPosition[1], 0.2f, 0.03f), false, false, 0, LEVELINFO_ATTEMPTS);
			TextPosition[1] += 0.03f;
			irrGUI->addStaticText(L"Wins", Interface::Instance().GetRightTextRect(L"Wins", TextPosition[0], TextPosition[1], TextPosition[1] + 0.03f), false, false, 0, -1);
			irrGUI->addStaticText(L"0", Interface::Instance().GetAbsoluteRectWH(TextPosition[0] + 0.02f, TextPosition[1], 0.2f, 0.03f), false, false, 0, LEVELINFO_WINS);
			TextPosition[1] += 0.03f;
			irrGUI->addStaticText(L"Losses", Interface::Instance().GetRightTextRect(L"Losses", TextPosition[0], TextPosition[1], TextPosition[1] + 0.03f), false, false, 0, -1);
			irrGUI->addStaticText(L"0", Interface::Instance().GetAbsoluteRectWH(TextPosition[0] + 0.02f, TextPosition[1], 0.2f, 0.03f), false, false, 0, LEVELINFO_LOSSES);
			TextPosition[1] += 0.03f;
			irrGUI->addStaticText(L"Play Time", Interface::Instance().GetRightTextRect(L"Play Time", TextPosition[0], TextPosition[1], TextPosition[1] + 0.03f), false, false, 0, -1);
			irrGUI->addStaticText(L"--", Interface::Instance().GetAbsoluteRectWH(TextPosition[0] + 0.02f, TextPosition[1], 0.2f, 0.03f), false, false, 0, LEVELINFO_PLAYTIME);
			TextPosition[1] += 0.03f;
			irrGUI->addStaticText(L"Best Time", Interface::Instance().GetRightTextRect(L"Best Time", TextPosition[0], TextPosition[1], TextPosition[1] + 0.03f), false, false, 0, -1);
			irrGUI->addStaticText(L"--", Interface::Instance().GetAbsoluteRectWH(TextPosition[0] + 0.02f, TextPosition[1], 0.2f, 0.03f), false, false, 0, LEVELINFO_BESTTIME);

			// Add level list
			const CampaignStruct &Campaign = Campaign::Instance().GetCampaign(CampaignIndex);
			float Position[2] = { 0.3f, 0.1f };

			bool SaveNeeded = false;
			for(u32 i = 0; i < Campaign.Levels.size(); i++) {
				
				// Get unlock status
				SaveLevelStruct Stats;
				Save::Instance().GetLevelStats(Campaign.Levels[i].File, Stats);
				if(Campaign.Levels[i].UnlockLevel > Stats.UnlockLevel) {
					Stats.UnlockLevel = Campaign.Levels[i].UnlockLevel;
					Save::Instance().UpdateLevelStats(Campaign.Levels[i].File, Stats);
					SaveNeeded = true;
				}

				// Add button
				IGUIButton *Level = irrGUI->addButton(Interface::Instance().GetAbsoluteRectWH(Position[0], Position[1], Interface::Instance().GetFractionX(64), Interface::Instance().GetFractionY(64)), 0, CAMPAIGN_LEVELID + i);

				if(Stats.UnlockLevel > 0)
					Level->setImage(irrDriver->getTexture((stringc("levels/icons/") + Campaign.Levels[i].Icon).c_str()));
				else
					Level->setImage(irrDriver->getTexture("levels/icons/locked.jpg"));

				Position[0] += 0.1f;
				if(Position[0] >= 0.9f) {
					Position[0] = 0.3f;
					Position[1] += 0.1f;
				}				
			}

			if(SaveNeeded)
				Save::Instance().SaveLevelStats();

			// Add level selector
			IGUIImage *SelectedImage = irrGUI->addImage(irrDriver->getTexture("levels/icons/selected.png"), position2di(0, 0), true, 0, LEVELS_SELECTEDLEVEL);

			// Confirmations
			IGUIButton *ButtonGo = irrGUI->addButton(Interface::Instance().GetAbsoluteRectCentered(0.2f, 0.9f, 0.1f, 0.05f), 0, LEVELS_GO, L"Go");
			IGUIButton *ButtonBuy = irrGUI->addButton(Interface::Instance().GetAbsoluteRectCentered(0.4f, 0.9f, 0.1f, 0.05f), 0, LEVELS_BUY, L"Shop");
			IGUIButton *ButtonHighscores = irrGUI->addButton(Interface::Instance().GetAbsoluteRectCentered(0.6f, 0.9f, 0.1f, 0.05f), 0, LEVELS_HIGHSCORES, L"Highscores");
			IGUIButton *ButtonBack = irrGUI->addButton(Interface::Instance().GetAbsoluteRectCentered(0.8f, 0.9f, 0.1f, 0.05f), 0, LEVELS_BACK, L"Back");
			ButtonBuy->setEnabled(false);
			ButtonHighscores->setEnabled(false);

			// Selected level
			SelectedLevel = 0;
			UpdateSelectedLevel(SelectedLevel);
			UpdateLevelInfo(SelectedLevel);

			State = STATE_LEVELS;
		}
		break;
		case STATE_LEVELS:
		break;
		case STATE_INITREPLAYS: {
			Interface::Instance().Clear();
			
			// Level selection
			IGUIListBox *ListReplays = irrGUI->addListBox(Interface::Instance().GetAbsoluteRect(0.1f, 0.2f, 0.9f, 0.75f), 0, REPLAYS_FILES, true);

			// Change directories
			stringc OldWorkingDirectory = irrFile->getWorkingDirectory();
			irrFile->changeWorkingDirectoryTo(Save::Instance().GetReplayPath().string().c_str());

			// Get a list of replays
			IFileList *FileList = irrFile->createFileList();
			int FileCount = FileList->getFileCount();
			for(int i = 0; i < FileCount; i++) {
				if(!FileList->isDirectory(i)) {
					stringw File(FileList->getFileName(i));

					ReplayHeaderStruct Header;
					Replay::Instance().GetReplayHeader(File.c_str(), Header);
					if(Header.ReplayVersion == REPLAY_VERSION && File.find(".replay") != -1) {
						stringw ReplayInfo = File;
						ReplayInfo += " - ";
						ReplayInfo += Header.LevelFile;
						ReplayInfo += " - ";
						ReplayInfo += Header.Description;
						ReplayInfo += " - ";
						ReplayInfo += Interface::Instance().ConvertSecondsToString(Header.CompletionTime);
						ListReplays->addItem(ReplayInfo.c_str());
					}
				}
			}
			irrFile->changeWorkingDirectoryTo(OldWorkingDirectory.c_str());

			// Confirmations
			IGUIButton *ButtonGo = irrGUI->addButton(Interface::Instance().GetAbsoluteRectCentered(0.3f, 0.8f, 0.15f, 0.05f), 0, REPLAYS_GO, L"Go");
			IGUIButton *ButtonDelete = irrGUI->addButton(Interface::Instance().GetAbsoluteRectCentered(0.5f, 0.8f, 0.15f, 0.05f), 0, REPLAYS_DELETE, L"Delete");
			IGUIButton *ButtonBack = irrGUI->addButton(Interface::Instance().GetAbsoluteRectCentered(0.7f, 0.8f, 0.15f, 0.05f), 0, REPLAYS_BACK, L"Back");

			State = STATE_REPLAYS;
		}
		break;
		case STATE_REPLAYS:
		break;
		case STATE_INITOPTIONS:	{
			Interface::Instance().Clear();
			IGUIButton *ButtonVideo = irrGUI->addButton(Interface::Instance().GetAbsoluteRect(0.4f, 0.3f, 0.6f, 0.35f), 0, OPTIONS_VIDEO, L"Video");
			IGUIButton *ButtonAudio = irrGUI->addButton(Interface::Instance().GetAbsoluteRect(0.4f, 0.4f, 0.6f, 0.45f), 0, OPTIONS_AUDIO, L"Audio");
			IGUIButton *ButtonControls = irrGUI->addButton(Interface::Instance().GetAbsoluteRect(0.4f, 0.5f, 0.6f, 0.55f), 0, OPTIONS_CONTROLS, L"Controls");
			IGUIButton *ButtonBack = irrGUI->addButton(Interface::Instance().GetAbsoluteRect(0.4f, 0.6f, 0.6f, 0.65f), 0, OPTIONS_BACK, L"Back");

			State = STATE_OPTIONS;
		}
		break;
		case STATE_OPTIONS:
		break;
		case STATE_INITVIDEO: {
			Interface::Instance().Clear();

			// Video modes
			const array<VideoModeStruct> &ModeList = Graphics::Instance().GetVideoModes();
			if(ModeList.size() > 0) {
				IGUIStaticText *TextScreenResolution = irrGUI->addStaticText(L"Screen Resolution", Interface::Instance().GetAbsoluteRect(0.3f, 0.2f, 0.5f, 0.23f), false, false);
				IGUIComboBox *ListScreenResolution = irrGUI->addComboBox(Interface::Instance().GetAbsoluteRect(0.5f, 0.2f, 0.7f, 0.23f), 0, VIDEO_VIDEOMODES);

				// Populate mode list
				for(u32 i = 0; i < ModeList.size(); i++)
					ListScreenResolution->addItem(ModeList[i].String.c_str());
				ListScreenResolution->setSelected(Graphics::Instance().GetCurrentVideoModeIndex());
			}
	
			// Full Screen
			IGUIStaticText *TextFullScreen = irrGUI->addStaticText(L"Full Screen", Interface::Instance().GetAbsoluteRect(0.3f, 0.25f, 0.5f, 0.28f), false, false);
			IGUICheckBox *CheckBoxFullScreen = irrGUI->addCheckBox(Config::Instance().FullScreen, Interface::Instance().GetAbsoluteRect(0.5f, 0.25f, 0.7f, 0.28f), 0, VIDEO_FULLSCREEN);

			// Save
			IGUIButton *ButtonSave = irrGUI->addButton(Interface::Instance().GetAbsoluteRect(0.35f, 0.6f, 0.49f, 0.65f), 0, VIDEO_SAVE, L"Save");
			IGUIButton *ButtonCancel = irrGUI->addButton(Interface::Instance().GetAbsoluteRect(0.51f, 0.6f, 0.65f, 0.65f), 0, VIDEO_CANCEL, L"Cancel");
			IGUIStaticText *TextWarning = irrGUI->addStaticText(L"Changes are applied after restart", Interface::Instance().GetCenteredTextRect(L"Changes are applied after restart", 0.70f, 0.75f), false, false, 0, -1, true);

			State = STATE_VIDEO;
		}
		break;
		case STATE_VIDEO:
		break;
		case STATE_INITAUDIO: {
			Interface::Instance().Clear();

			// Audio enabled
			IGUIStaticText *TextAudioEnabled = irrGUI->addStaticText(L"Audio Enabled", Interface::Instance().GetAbsoluteRectWH(0.3f, 0.25f, 0.2f, 0.03f), false, false);
			IGUICheckBox *CheckBoxAudioEnabled = irrGUI->addCheckBox(Config::Instance().AudioEnabled, Interface::Instance().GetAbsoluteRectWH(0.5f, 0.25f, 0.2f, 0.03f), 0, AUDIO_ENABLED);

			IGUIStaticText *TextSoundVolume= irrGUI->addStaticText(L"Sound Volume", Interface::Instance().GetAbsoluteRectWH(0.3f, 0.35f, 0.2f, 0.03f), false, false);
			IGUIScrollBar *ScrollBarSoundVolume = irrGUI->addScrollBar(1, Interface::Instance().GetAbsoluteRectWH(0.5f, 0.35f, 0.3f, 0.05f), 0, AUDIO_SOUND);
			ScrollBarSoundVolume->setMax(100);
			ScrollBarSoundVolume->setPos((int)(Config::Instance().SoundVolume * 100));

			IGUIStaticText *TextMusicVolume= irrGUI->addStaticText(L"Music Volume", Interface::Instance().GetAbsoluteRectWH(0.3f, 0.45f, 0.2f, 0.03f), false, false);
			IGUIScrollBar *ScrollBarMusicVolume = irrGUI->addScrollBar(1, Interface::Instance().GetAbsoluteRectWH(0.5f, 0.45f, 0.3f, 0.05f), 0, AUDIO_MUSIC);
			ScrollBarMusicVolume->setMax(100);
			ScrollBarMusicVolume->setPos((int)(Config::Instance().MusicVolume * 100));

			// Save
			IGUIButton *ButtonSave = irrGUI->addButton(Interface::Instance().GetAbsoluteRect(0.35f, 0.9f, 0.49f, 0.95f), 0, AUDIO_SAVE, L"Save");
			IGUIButton *ButtonCancel = irrGUI->addButton(Interface::Instance().GetAbsoluteRect(0.51f, 0.9f, 0.65f, 0.95f), 0, AUDIO_CANCEL, L"Cancel");

			State = STATE_AUDIO;
		}
		break;
		case STATE_AUDIO:
		break;
		case STATE_INITCONTROLS: {
			Interface::Instance().Clear();
			KeyButton = NULL;

			// Create the key buttons
			float StartX = 0.1f, StartY = 0.1f;
			for(int i = 0; i < ACTIONS::COUNT; i++) {
				
				CurrentKeys[i] = Config::Instance().Keys[i];
				irrGUI->addStaticText(ActionNames[i], Interface::Instance().GetAbsoluteRectWH(StartX, StartY, 0.2f, 0.05f), false, false);
				irrGUI->addButton(Interface::Instance().GetAbsoluteRectWH(StartX + 0.2f, StartY, 0.13f, 0.05f), 0, CONTROLS_MOVEFORWARD + i, stringw(Input::Instance().GetKeyName(CurrentKeys[i])).c_str());

				StartY += 0.06f;
			}

			// Save
			IGUIButton *ButtonSave = irrGUI->addButton(Interface::Instance().GetAbsoluteRect(0.35f, 0.9f, 0.49f, 0.95f), 0, CONTROLS_SAVE, L"Save");
			IGUIButton *ButtonCancel = irrGUI->addButton(Interface::Instance().GetAbsoluteRect(0.51f, 0.9f, 0.65f, 0.95f), 0, CONTROLS_CANCEL, L"Cancel");

			State = STATE_CONTROLS;
		}
		break;
		case STATE_CONTROLS:
		break;
	}
}

// Draws the current state
void MenuState::Draw() {
	
}

// Cancels the key bind state
void MenuState::CancelKeyBind() {
	KeyButton->setText(KeyButtonOldText.c_str());
	KeyButton = NULL;
}

// Gets the replay name from a selection box
stringc MenuState::GetReplayFile() {

	// Get list
	IGUIListBox *ReplayList = static_cast<IGUIListBox *>(irrGUI->getRootGUIElement()->getElementFromId(REPLAYS_FILES));
	
	int SelectedIndex = ReplayList->getSelected();
	if(SelectedIndex != -1) {

		// Get replay file name
		stringc ListItem = ReplayList->getListItem(SelectedIndex);
		int SpaceIndex = ListItem.findFirstChar(" ", 1);

		return ListItem.subString(0, SpaceIndex);
	}

	return "";
}

// Launchs a level
void MenuState::LaunchLevel() {
	
	SaveLevelStruct Stats;
	Save::Instance().GetLevelStats(Campaign::Instance().GetCampaign(CampaignIndex).Levels[SelectedLevel].File, Stats);
	if(Stats.UnlockLevel == 0)
		return;

	Campaign::Instance().SetCurrentCampaign(CampaignIndex);
	Campaign::Instance().SetCurrentLevel(SelectedLevel);
	PlayState::Instance()->SetNextLevel(Campaign::Instance().GetLevel());
	Game::Instance().ChangeState(PlayState::Instance());
}

// Launchs a replay from a list item
void MenuState::LaunchReplay() {

	// Get replay file
	stringc File = GetReplayFile();
	if(File != "") {

		// Load replay
		ViewReplayState::Instance()->SetCurrentReplay(File);
		Game::Instance().ChangeState(ViewReplayState::Instance());
	}
}

// Updates the selected level image
void MenuState::UpdateSelectedLevel(int TIndex) {

	IGUIImage *SelectedImage = static_cast<IGUIImage *>(irrGUI->getRootGUIElement()->getElementFromId(LEVELS_SELECTEDLEVEL));
	IGUIButton *SelectedLevel = static_cast<IGUIButton *>(irrGUI->getRootGUIElement()->getElementFromId(CAMPAIGN_LEVELID + TIndex));

	SelectedImage->setRelativePosition(SelectedLevel->getAbsolutePosition());
}

// Updates the level info box
void MenuState::UpdateLevelInfo(int TIndex) {
	
	const CampaignStruct &Campaign = Campaign::Instance().GetCampaign(CampaignIndex);
	if(TIndex < 0 || TIndex >= (int)Campaign.Levels.size())
		return;

	// Get level stats
	SaveLevelStruct Stats;
	Save::Instance().GetLevelStats(Campaign.Levels[TIndex].File, Stats);

	IGUIStaticText *TextDescription = static_cast<IGUIStaticText *>(irrGUI->getRootGUIElement()->getElementFromId(LEVELINFO_DESCRIPTION));
	IGUIStaticText *TextAttempts = static_cast<IGUIStaticText *>(irrGUI->getRootGUIElement()->getElementFromId(LEVELINFO_ATTEMPTS));
	IGUIStaticText *TextWins = static_cast<IGUIStaticText *>(irrGUI->getRootGUIElement()->getElementFromId(LEVELINFO_WINS));
	IGUIStaticText *TextLosses = static_cast<IGUIStaticText *>(irrGUI->getRootGUIElement()->getElementFromId(LEVELINFO_LOSSES));
	IGUIStaticText *TextPlayTime = static_cast<IGUIStaticText *>(irrGUI->getRootGUIElement()->getElementFromId(LEVELINFO_PLAYTIME));
	IGUIStaticText *TextBestTime = static_cast<IGUIStaticText *>(irrGUI->getRootGUIElement()->getElementFromId(LEVELINFO_BESTTIME));

	IGUIButton *ButtonGo = static_cast<IGUIButton *>(irrGUI->getRootGUIElement()->getElementFromId(LEVELS_GO));
	if(Stats.UnlockLevel == 0) {
		TextDescription->setText(L"Level Locked");
		TextAttempts->setText(L"0");
		TextWins->setText(L"0");
		TextLosses->setText(L"0");
		TextPlayTime->setText(L"--");
		TextBestTime->setText(L"--");

		ButtonGo->setEnabled(false);
	}
	else {
		TextDescription->setText(stringw(Campaign.Levels[TIndex].File.c_str()).c_str());
		TextAttempts->setText(stringw(Stats.LoadCount).c_str());
		TextWins->setText(stringw(Stats.WinCount).c_str());
		TextLosses->setText(stringw(Stats.LoseCount).c_str());
		TextPlayTime->setText(Interface::Instance().ConvertSecondsToString(Stats.PlayTime).c_str());
		if(Stats.HighScores.size() > 0)
			TextBestTime->setText(Interface::Instance().ConvertSecondsToString(Stats.HighScores[0].Time).c_str());
		else
			TextBestTime->setText(L"--");

		ButtonGo->setEnabled(true);
	}
}

// Reloads the menu audio files
void MenuState::ReloadAudio() {
	Audio::Instance().Load("menu_click", "sounds/menu_click.ogg", false);
	Audio::Instance().Load("menu_cancel", "sounds/menu_cancel.ogg", false);
}
