/*************************************************************************************
*	irrlamb - http://code.google.com/p/irrlamb/
*	Copyright (C) 2007  Alan Witkowski
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation, either version 3 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program.  If not, see <http://www.gnu.org/licenses/>.
**************************************************************************************/
#include "orb.h"
#include "../engine/globals.h"
#include "../engine/physics.h"
#include "../engine/scripting.h"
#include "../engine/game.h"
#include "../engine/replay.h"
#include "default.h"

// Constructor
OrbClass::OrbClass(DefaultObjectStruct &TObject)
:	RigidBodyClass(TObject),
	DeactivationCallback(""),
	DeactivationTime(-1),
	DeactivationLength(0),
	Deactivated(false),
	Deactivating(false) {

	// Graphics
	Node = irrScene->addSphereSceneNode(TObject.Radius, 32);
	if(TObject.Texture != "")
		Node->setMaterialTexture(0, irrDriver->getTexture(TObject.Texture.c_str()));
	else
		Node->setMaterialTexture(0, irrDriver->getTexture("textures/orb0.png"));
	Node->setMaterialType(EMT_TRANSPARENT_ADD_COLOR);
	
	// Create glow
	GlowColor = SColor(255, 255, 240, 130);
	InnerNode = irrScene->addBillboardSceneNode(Node, dimension2df(4.0f, 4.0f));
	InnerNode->setColor(GlowColor);
	InnerNode->setMaterialFlag(EMF_LIGHTING, false);
	InnerNode->setMaterialFlag(EMF_ZBUFFER, false);
	InnerNode->setMaterialType(EMT_TRANSPARENT_ADD_COLOR);
	if(TObject.DetailTexture != "")
		InnerNode->setMaterialTexture(0, irrDriver->getTexture(TObject.DetailTexture.c_str()));
	else
		InnerNode->setMaterialTexture(0, irrDriver->getTexture("textures/glow0.png"));

	if(Physics::Instance().IsEnabled()) {
	
		// Orientation
		btTransform Transform;
		Physics::Instance().GetBulletTransform(Transform, TObject.Position, TObject.Rotation);
		MotionState = new btDefaultMotionState(Transform);
		
		// Create shape
		Shape = new btSphereShape(TObject.Radius);

		btVector3 LocalInertia;
		Shape->calculateLocalInertia(TObject.Mass, LocalInertia);

		RigidBody = new btRigidBody(TObject.Mass, MotionState, Shape, LocalInertia);
		RigidBody->setUserPointer((void *)this);

		Physics::Instance().GetWorld()->addRigidBody(RigidBody, TObject.CollisionGroup, TObject.CollisionMask);
	}

	SetProperties(TObject);
}

// Destructor
OrbClass::~OrbClass() {
	
	Node->remove();
}

// Deactivates the object
void OrbClass::StartDeactivation(const stringc &TCallback, u32 TTime) {

	if(!Deactivated && !Deactivating) {
		Deactivating = true;
		DeactivationCallback = TCallback;
		DeactivationLength = TTime;
		DeactivationTime = Timer.GetTime() + DeactivationLength;

		// Save the event on the replay
		if(Replay::Instance().IsRecording()) {
			Replay::Instance().WriteEvent(ReplayClass::PACKET_ORBDEACTIVATE);

			ReplayOrbDeactivateStruct Packet;
			Packet.ObjectID = ID;
			Packet.Length = TTime;
			Replay::Instance().WritePacket(Packet);
		}
	}
}

// Updates the orb
void OrbClass::Update(u32 TDeltaTime) {

	RigidBodyClass::Update(TDeltaTime);

	UpdateDeactivation();
}

// Updates the replay
void OrbClass::UpdateReplay(u32 TDeltaTime) {
	RigidBodyClass::UpdateReplay(TDeltaTime);

	UpdateDeactivation();
}

// Updates the deactivation
void OrbClass::UpdateDeactivation() {

	if(Deactivating) {
		if(Timer.GetTime() >= DeactivationTime) {
			Scripting::Instance().CallFunction(DeactivationCallback);
			InnerNode->setVisible(false);
			Deactivating = false;
			Deactivated = true;
		}
		else {
			float PercentLeft = (DeactivationTime - Timer.GetTime()) / (float)DeactivationLength;
			InnerNode->setSize(dimension2df(4.0f * PercentLeft, 4.0f * PercentLeft));
		}
	}
}

