/*************************************************************************************
*	irrlamb - http://code.google.com/p/irrlamb/
*	Copyright (C) 2007  Alan Witkowski
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation, either version 3 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program.  If not, see <http://www.gnu.org/licenses/>.
**************************************************************************************/
#include "player.h"
#include "../engine/globals.h"
#include "../engine/physics.h"
#include "../engine/objectmanager.h"
#include "../engine/input.h"
#include "../engine/config.h"
#include "../engine/game.h"
#include "../engine/audio.h"
#include "../engine/level.h"
#include "sphere.h"
#include "basicjoint.h"
#include "springjoint.h"
#include "default.h"

// Constructor
PlayerClass::PlayerClass(DefaultObjectStruct &TObject)
:	RigidBodyClass(TObject),
	Camera(NULL),
	JumpTimer(0),
	GrappleState(GRAPPLE_READY),
	GrappleBody(NULL),
	GrappleRequestState(GRAPPLEREQUEST_NONE),
	StatGrappleLength(100.0f),
	GrappleEnabled(TObject.GrappleEnabled) {

	TorqueFactor = TObject.TorqueFactor;
	Radius = TObject.Radius;

	// Graphics
	Node = irrScene->addSphereSceneNode(Radius);
	if(TObject.Texture != "")
		Node->setMaterialTexture(0, irrDriver->getTexture(TObject.Texture.c_str()));
	else
		Node->setMaterialTexture(0, irrDriver->getTexture("textures/player0.png"));
	Node->setMaterialType(EMT_TRANSPARENT_ADD_COLOR);
	
	// Add glow
	ISceneNode *InnerNode;
	InnerNode = irrScene->addBillboardSceneNode(Node, dimension2df(2.2f, 2.2f));
	InnerNode->setMaterialFlag(EMF_LIGHTING, false);
	InnerNode->setMaterialFlag(EMF_ZBUFFER, false);
	InnerNode->setMaterialType(EMT_TRANSPARENT_ADD_COLOR);
	InnerNode->setMaterialTexture(0, irrDriver->getTexture("textures/fire1.png"));

	// Particle system
	IParticleSystemSceneNode *ParticleSystem = irrScene->addParticleSystemSceneNode(false, Node);
	ParticleSystem->setParticleSize(dimension2d<f32>(0.25f, 0.25f));

	// Create the emitter
	IParticleEmitter *Emitter = ParticleSystem->createBoxEmitter(
		aabbox3df(-0.05f, -0.05f, -0.05f, 0.05f, 0.05f, 0.05f), 
		vector3df(0.0f, 0.0f, 0.0f),
		60, 80,
		SColor(0, 255, 255, 255), video::SColor(0, 255, 255, 255),
		1000, 1000);
	ParticleSystem->setEmitter(Emitter);
	Emitter->drop();

	// Make the particles fade out
	IParticleAffector *ParticleAffector = ParticleSystem->createFadeOutParticleAffector();
	ParticleSystem->addAffector(ParticleAffector);
	ParticleAffector->drop();

	// Set materials
	ParticleSystem->setMaterialFlag(EMF_LIGHTING, false);
	ParticleSystem->setMaterialFlag(EMF_ZBUFFER, false);
	ParticleSystem->setMaterialTexture(0, irrDriver->getTexture("textures/fire1.png"));
	ParticleSystem->setMaterialType(EMT_TRANSPARENT_VERTEX_ALPHA);

	if(Physics::Instance().IsEnabled()) {

		// Orientation
		btTransform Transform;
		Physics::Instance().GetBulletTransform(Transform, TObject.Position, TObject.Rotation);
		MotionState = new btDefaultMotionState(Transform);

		Shape = new btSphereShape(static_cast<btScalar>(TObject.Radius));

		// Calculate inertia
		btVector3 LocalInertia;
		Shape->calculateLocalInertia(TObject.Mass, LocalInertia);

		RigidBody = new btRigidBody(TObject.Mass, MotionState, Shape, LocalInertia);
		RigidBody->setUserPointer((void *)this);

		Physics::Instance().GetWorld()->addRigidBody(RigidBody, TObject.CollisionGroup, TObject.CollisionMask);
		RigidBody->setActivationState(DISABLE_DEACTIVATION);
	}

	SetProperties(TObject);
	Hookable = false;
	if(CollisionCallback == "")
		CollisionCallback = "HitPlayer";
}

// Destructor
PlayerClass::~PlayerClass() {

	Node->remove();
}

// Update the player
void PlayerClass::Update(u32 TDeltaTime) {
	RigidBodyClass::Update(TDeltaTime);

	JumpTimer -= TDeltaTime;
	if(JumpTimer < 0)
		JumpTimer = 0;

	UpdateGrapple();
}

// Processes key presses
bool PlayerClass::ProcessKeyPress(int TKey) {

	if(TKey == Config::Instance().Keys[ACTIONS::JUMP]) {
		if(TouchingGround && JumpTimer == 0) {
			RigidBody->applyCentralImpulse(btVector3(0.0f, 5.0f, 0.0f));
			JumpTimer = 15;
		}
	}
	
	else
		return false;
	
	return true;
}

// Handles a mouse press
bool PlayerClass::HandleMousePress(int TButton, int TMouseX, int TMouseY) {

	switch(TButton) {
		// Right button
		case 1:
			if(GrappleEnabled)
				GrappleRequestState = GRAPPLEREQUEST_SHOOT;
		break;
		default:
			return false;
		break;
	}

	return true;
}

// Handles a mouse lift
bool PlayerClass::HandleMouseLift(int TButton, int TMouseX, int TMouseY) {

	switch(TButton) {
		// Right button
		case 1:
			if(GrappleEnabled)
				GrappleRequestState = GRAPPLEREQUEST_RELEASE;
		break;
		default:
			return false;
		break;
	}

	return true;
}

// Processes input from the keyboard
void PlayerClass::HandleInput() {
	
	// Move object
	vector3df Push(0.0f, 0.0f, 0.0f);
	if(Input::Instance().GetKeyState(Config::Instance().Keys[ACTIONS::MOVEFORWARD]))
		Push.Z = 1.0f;
	else if(Input::Instance().GetKeyState(Config::Instance().Keys[ACTIONS::MOVEBACK]))
		Push.Z = -1.0f;

	if(Input::Instance().GetKeyState(Config::Instance().Keys[ACTIONS::MOVELEFT]))
		Push.X = -1.0f;
	else if(Input::Instance().GetKeyState(Config::Instance().Keys[ACTIONS::MOVERIGHT]))
		Push.X = 1.0f;

	if(!iszero(Push.getLengthSQ())) {
		Push.normalize();

		matrix4 DirectionTransform;
		DirectionTransform.makeIdentity();
		DirectionTransform.setRotationDegrees(vector3df(0.0f, Camera->GetYaw(), 0.0f)); 
		DirectionTransform.transformVect(Push);

		vector3df RotationAxis = Push.crossProduct(vector3df(0.0f, -1.0f, 0.0f));

		RotationAxis *= (float)(TorqueFactor * Game::Instance().GetDeltaTimeSeconds() * Physics::Instance().GetSimulationSpeed());
		RigidBody->applyTorque(btVector3(RotationAxis.X, RotationAxis.Y, RotationAxis.Z));
	}

}

// Stop the grapple and create a hook
void PlayerClass::GrappleTipHit(RigidBodyClass *TTip, RigidBodyClass *TObject) {
	
	
	// Get template
	DefaultObjectStruct GrappleFixed;
	int GrappleFixedIndex = Level::Instance().GetCustomObjectTemplateIndex(LevelClass::CUSTOM_GRAPPLEFIXED);
	Level::Instance().GetTemplateFromIndex(GrappleFixedIndex, GrappleFixed);

	// Create the fixed joint
	GrappleFixed.BodyA = TObject;
	GrappleFixed.BodyB = TTip;
	GrappleJoints.push_back(ObjectManager::Instance().AddObject(new BasicJointClass(GrappleFixed)));
	Level::Instance().WriteCreateEvent(GrappleFixedIndex);

	// Get template
	DefaultObjectStruct GrappleSpring;
	int GrappleSpringIndex = Level::Instance().GetCustomObjectTemplateIndex(LevelClass::CUSTOM_GRAPPLESPRING);
	Level::Instance().GetTemplateFromIndex(GrappleSpringIndex, GrappleSpring);

	// Create the spring joint
	GrappleSpring.BodyA = TTip;
	GrappleSpring.BodyB = this;
	GrappleJoints.push_back(ObjectManager::Instance().AddObject(new SpringJointClass(GrappleSpring)));
	Level::Instance().WriteCreateEvent(GrappleSpringIndex);

	Audio::Instance().Play("game_grapplehold");
	GrappleState = GRAPPLE_HOOKED;
}

// Updates the grapple state
void PlayerClass::UpdateGrapple() {

	switch(GrappleState) {
		case GRAPPLE_READY:
			if(GrappleRequestState == GRAPPLEREQUEST_SHOOT) {
				GrappleState = GRAPPLE_SHOOT;
				GrappleRequestState = GRAPPLEREQUEST_NONE;
			}
		break;
		case GRAPPLE_SHOOT: {

			DefaultObjectStruct GrappleProperties;
			Level::Instance().GetTemplateFromIndex(Level::Instance().GetCustomObjectTemplateIndex(LevelClass::CUSTOM_GRAPPLETIP), GrappleProperties);

			// Set Position
			vector3df GrapplePosition(RigidBody->getWorldTransform().getOrigin()[0], RigidBody->getWorldTransform().getOrigin()[1], RigidBody->getWorldTransform().getOrigin()[2]);
			Level::Instance().SetDefaultPosition(GrapplePosition);

			// Create object
			GrappleBody = static_cast<GrappleClass *>(Level::Instance().CreateObjectFromTemplateIndex(Level::Instance().GetCustomObjectTemplateIndex(LevelClass::CUSTOM_GRAPPLETIP)));
			GrappleBody->SetGrappleParent(this);
			
			// Set direction
			btVector3 GrappleVelocity(Camera->GetDirection().X, Camera->GetDirection().Y, Camera->GetDirection().Z);
			GrappleVelocity *= 20.0f;
			GrappleBody->GetBody()->setLinearVelocity(GrappleVelocity + RigidBody->getLinearVelocity());

			//Audio::Instance().Play("game_grappleshoot");
			GrappleState = GRAPPLE_FLYING;
		}
		break;
		case GRAPPLE_FLYING:
			if(GrappleRequestState == GRAPPLEREQUEST_RELEASE || !CheckGrappleDistance()) {
				GrappleState = GRAPPLE_DELETECHAIN;
				GrappleRequestState = GRAPPLEREQUEST_NONE;
			}
		break;
		case GRAPPLE_HOOKED:
			if(GrappleRequestState == GRAPPLEREQUEST_RELEASE) {
				Audio::Instance().Stop("game_grapplehold");
				GrappleState = GRAPPLE_DELETEJOINT;
				GrappleRequestState = GRAPPLEREQUEST_NONE;
			}
		break;
		case GRAPPLE_DELETEJOINT:

			// Remove the joint
			for(list<ObjectClass *>::Iterator Iterator = GrappleJoints.begin(); Iterator != GrappleJoints.end(); ++Iterator) {
				(*Iterator)->SetDeleted(true);
			}
			GrappleJoints.clear();

			GrappleState = GRAPPLE_DELETECHAIN;
		break;
		case GRAPPLE_DELETECHAIN:
			Audio::Instance().Play("game_grapplebreak");
			GrappleBody->SetDeleted(true);
			GrappleBody = NULL;

			GrappleState = GRAPPLE_READY;
		break;
	}
}

// Returns false if the grapple is too far away from the player
bool PlayerClass::CheckGrappleDistance() {

	if(GrappleBody && (RigidBody->getWorldTransform().getOrigin() - GrappleBody->GetBody()->getWorldTransform().getOrigin()).length2() > StatGrappleLength)
		return false;

	return true;
}
