/*************************************************************************************
*	irrlamb - http://code.google.com/p/irrlamb/
*	Copyright (C) 2007  Alan Witkowski
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation, either version 3 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program.  If not, see <http://www.gnu.org/licenses/>.
**************************************************************************************/
#include "rigidbody.h"
#include "default.h"
#include "../engine/scripting.h"
#include "../engine/physics.h"
#include "../engine/replay.h"

// Constructor
RigidBodyClass::RigidBodyClass(DefaultObjectStruct &TObject)
:	ObjectClass(),
	TouchingGround(false),
	TouchingWall(false),
	MovementChanged(false) 
{

	if(TObject.Mass == 0) {
		Physics::Instance().SetBodyType(TObject.CollisionGroup, PhysicsClass::FILTER_STATIC);
		Physics::Instance().RemoveFilter(TObject.CollisionMask, PhysicsClass::FILTER_STATIC);
	}

}

// Destructor
RigidBodyClass::~RigidBodyClass() {
	
	if(Physics::Instance().IsEnabled()) {
		Physics::Instance().GetWorld()->removeRigidBody(RigidBody);
		delete Shape;
		delete MotionState;
		delete RigidBody;
	}
}

// Update the rigid body
void RigidBodyClass::Update(u32 TDeltaTime) {
	ObjectClass::Update(TDeltaTime);
}

// Updates the body while replaying
void RigidBodyClass::UpdateReplay(u32 TDeltaTime) {
	ObjectClass::UpdateReplay(TDeltaTime);
}

// Updates the graphical representation of the object
void RigidBodyClass::UpdateRender() {

	// Position
	btVector3 Point = RigidBody->getCenterOfMassPosition();
	Node->setPosition(vector3df((f32)Point[0], (f32)Point[1], (f32)Point[2]));
	
	// Rotation
	btVector3 EulerRotation;
	btQuaternion RigidRotation = RigidBody->getOrientation();
	Physics::Instance().QuaternionToEuler(RigidRotation, EulerRotation);
	Node->setRotation(vector3df(EulerRotation[0], EulerRotation[1], EulerRotation[2]));

	// Note changes
	if(!MovementChanged && (PreviousPosition != Node->getPosition() || PreviousRotation != Node->getRotation()))
		MovementChanged = true;

	PreviousPosition = Node->getPosition();
	PreviousRotation = Node->getRotation();
}

// Stops the body's movement
void RigidBodyClass::Stop() {

	RigidBody->setLinearVelocity(btVector3(0.0f, 0.0f, 0.0f));
	RigidBody->setAngularVelocity(btVector3(0.0f, 0.0f, 0.0f));
}

// Sets the position of the body
void RigidBodyClass::SetPosition(const btVector3 &TPosition) {

	RigidBody->getWorldTransform().setOrigin(TPosition);
}

// Returns the position of the body
btVector3 RigidBodyClass::GetPosition() {

	return RigidBody->getWorldTransform().getOrigin();
}

// Sets the body's linear velocity
void RigidBodyClass::SetLinearVelocity(const btVector3 &TPosition) {
	
	RigidBody->setLinearVelocity(TPosition);
}

// Returns the body's linear velocity
btVector3 RigidBodyClass::GetLinearVelocity() {

	return RigidBody->getLinearVelocity();
}

// Set common rigid body default values
void RigidBodyClass::SetProperties(DefaultObjectStruct &TObject) {

	ObjectClass::SetProperties(TObject);

	// Graphics
	Node->setPosition(TObject.Position);
	Node->setRotation(TObject.Rotation);
	Node->setMaterialFlag(EMF_LIGHTING, TObject.Lighting);
	Node->setMaterialFlag(EMF_FOG_ENABLE, TObject.Fog);
	Node->setMaterialFlag(EMF_NORMALIZE_NORMALS, true);
	Node->setVisible(TObject.Visible);

	// Physics
	if(Physics::Instance().IsEnabled()) {
		RigidBody->setFriction(TObject.Friction);
		RigidBody->setDamping(TObject.LinearDamping, TObject.AngularDamping);

		// Collision
		CollisionCallback = TObject.CollisionCallback;
		Hookable = TObject.Hookable;
	}
}

// Collision callback
void RigidBodyClass::HandleCollision(ObjectClass *TOtherObject, const btPersistentManifold *TContactManifold, float TNormalScale) {
	
	// Get touching states
	if(TOtherObject->GetType() != ZONE) {
		for(int i = 0; i < TContactManifold->getNumContacts(); i++) {
			float NormalY = TContactManifold->getContactPoint(i).m_normalWorldOnB[1] * TNormalScale;
			if(NormalY > 0.6f)
				TouchingGround = true;
			if(NormalY < 0.7f && NormalY > -0.7f)
				TouchingWall = true;
		}
	}

	if(CollisionCallback.size())
		Scripting::Instance().CallCollisionHandler(CollisionCallback, this, TOtherObject);
}

// Resets the object state before the frame begins
void RigidBodyClass::BeginFrame() {

	TouchingGround = TouchingWall = false;
}

// Fills out the replay packet with position and rotation
bool RigidBodyClass::GetReplayPacket(ReplayMovementStruct *TPacket) {

	if(MovementChanged) {
		TPacket->Position = Node->getPosition();
		TPacket->Rotation = Node->getRotation();
		MovementChanged = false;
		return true;
	}

	return false;
}

