/*************************************************************************************
*	irrlamb - http://code.google.com/p/irrlamb/
*	Copyright (C) 2007  Alan Witkowski
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation, either version 3 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program.  If not, see <http://www.gnu.org/licenses/>.
**************************************************************************************/
#include "play.h"
#include "engine/globals.h"
#include "engine/input.h"
#include "engine/graphics.h"
#include "engine/config.h"
#include "engine/physics.h"
#include "engine/scripting.h"
#include "engine/objectmanager.h"
#include "engine/replay.h"
#include "engine/audio.h"
#include "engine/interface.h"
#include "engine/random.h"
#include "engine/camera.h"
#include "engine/game.h"
#include "engine/level.h"
#include "engine/save.h"
#include "objects/player.h"
#include "menu.h"
#include "viewreplay.h"

PlayState::PlayState()
:	Camera(NULL) {

}

// Initializes the state
int PlayState::Init() {

	Timer.Start();
	InitPlay();

	// Set up physics world
	Physics::Instance().Init(true);
	Physics::Instance().SetSimulationSpeed(1.5);

	// Add camera
	Camera = new CameraClass();
	Camera->SetRotation(0.0f, 30.0f);
	Camera->SetDistance(5.0f);

	// Load the level
	CurrentLevel = NextLevel;
	NextLevel = "";
	if(!Level::Instance().Init(CurrentLevel, true))
		return 0;

	// Get the player
	Player = static_cast<PlayerClass *>(ObjectManager::Instance().GetObjectByName("player0"));
	if(Player == NULL)
		return 0;
	Player->SetCamera(Camera);

	// Handle saves
	Save::Instance().IncrementLevelLoadCount(CurrentLevel);

	Audio::Instance().Load("game_orb", "sounds/game_orb.ogg", false, 1.0f, true);
	Audio::Instance().Load("game_deactivate", "sounds/game_deactivate.ogg", false, 1.0f);
	Audio::Instance().Load("game_grappleshoot", "sounds/game_grappleshoot.ogg", false, 1.0f);
	Audio::Instance().Load("game_grapplehold", "sounds/game_grapplehold.ogg", false, 1.0f, true);
	Audio::Instance().Load("game_grapplebreak", "sounds/game_grapplebreak.ogg", false, 1.0f);

	return 1;
}

// Shuts the state down
int PlayState::Close() {

	// Stop the replay
	Replay::Instance().StopRecording();

	// Close the system down
	delete Camera;
	Level::Instance().Close();
	Interface::Instance().Clear();
	Physics::Instance().Close();
	irrScene->clear();

	Save::Instance().IncrementLevelPlayTime(CurrentLevel, Timer.GetTime());
	Save::Instance().SaveLevelStats();

	Audio::Instance().StopAll();
	
	return 1;
}

// Key presses
bool PlayState::HandleKeyPress(EKEY_CODE TKey) {

	bool Processed = true, LuaProcessed = false;
	
	switch(State) {
		case STATE_PLAY:

			if(TKey == Config::Instance().Keys[ACTIONS::RESET]) {
				NextLevel = CurrentLevel;
				Game::Instance().ChangeState(this);
			}

			switch(TKey) {
				case KEY_ESCAPE:
					if(Game::Instance().GetQuitOnEscape())
						Game::Instance().SetDone(true);
					else
						InitPause();
				break;
				case KEY_F1:
					InitPause();
				break;
				case KEY_F2:
					//InitWin();
				break;
				case KEY_F12:
					Graphics::Instance().SaveScreenshot();					
				break;
				default:
					Processed = Player->ProcessKeyPress(TKey);
				break;
			}

			// Send key presses to Lua
			LuaProcessed = Scripting::Instance().HandleKeyPress(TKey);
		break;
		case STATE_PAUSED:
			switch(TKey) {
				case KEY_ESCAPE:
					InitPlay();
				break;
			}
		break;
		case STATE_SAVEREPLAY:
			switch(TKey) {
				case KEY_ESCAPE:
					InitPause();
				break;
				case KEY_RETURN:
					SaveReplay();
				break;
				default:
					Processed = false;
				break;
			}
		break;
		case STATE_LOSE:

			if(TKey == Config::Instance().Keys[ACTIONS::RESET])
				Game::Instance().ChangeState(this);

			switch(TKey) {
				case KEY_ESCAPE:
					Game::Instance().ChangeState(MenuState::Instance());
				break;
				default:
					Processed = false;
				break;
			}
		break;
		case STATE_WIN:
			switch(TKey) {
				case KEY_ESCAPE:
					Game::Instance().ChangeState(MenuState::Instance());
				break;
				default:
					Processed = false;
				break;
			}
		break;
	}

	return Processed || LuaProcessed;
}

// Key lifts
bool PlayState::HandleKeyLift(EKEY_CODE TKey) {

	return false;
}

// Mouse motion
void PlayState::HandleMouseMotion(float TUpdateX, float TUpdateY) {

	switch(State) {
		case STATE_PLAY:
			Camera->HandleMouseMotion(TUpdateX, TUpdateY);
		break;
	}
}

// Mouse buttons
bool PlayState::HandleMousePress(int TButton, int TMouseX, int TMouseY) {

	switch(State) {
		case STATE_PLAY:
			Scripting::Instance().HandleMousePress(TButton, TMouseX, TMouseY); 
			Player->HandleMousePress(TButton, TMouseX, TMouseY);
		break;
	}

	return false;
}

// Mouse buttons
void PlayState::HandleMouseLift(int TButton, int TMouseX, int TMouseY) {
	Player->HandleMouseLift(TButton, TMouseX, TMouseY);
}

// Mouse wheel
void PlayState::HandleMouseWheel(float TDirection) {

}

// GUI events
void PlayState::HandleGUI(EGUI_EVENT_TYPE TEventType, IGUIElement *TElement) {

	switch(TEventType) {
		case EGET_BUTTON_CLICKED:
			switch(TElement->getID()) {
				case PAUSE_RESUME:
					InitPlay();
				break;
				case PAUSE_SAVEREPLAY:
					TargetState = STATE_PAUSED;
					InitSaveReplay();
				break;
				case PAUSE_MAINMENU:
					Game::Instance().ChangeState(MenuState::Instance());
				break;
				case SAVEREPLAY_SAVE:
					SaveReplay();
				break;
				case SAVEREPLAY_CANCEL:
					if(TargetState == STATE_WIN)
						InitWin();
					else
						InitPause();
				break;
				case LOSE_RESTARTLEVEL:
					Game::Instance().ChangeState(PlayState::Instance());
				break;
				case LOSE_SAVEREPLAY:
					TargetState = STATE_LOSE;
					InitSaveReplay();
				break;
				case LOSE_MAINMENU:
					Game::Instance().ChangeState(MenuState::Instance());
				break;
				case WIN_RESTARTLEVEL:
					NextLevel = CurrentLevel;
					Game::Instance().ChangeState(PlayState::Instance());
				break;
				case WIN_NEXTLEVEL:
					Game::Instance().ChangeState(PlayState::Instance());
				break;
				case WIN_SAVEREPLAY:
					TargetState = STATE_WIN;
					InitSaveReplay();
				break;
				case WIN_MAINMENU:
					Game::Instance().ChangeState(MenuState::Instance());
				break;
			}
		break;
	}
}

// Updates the current state
void PlayState::Update(u32 TDeltaTime) {

	switch(State) {
		case STATE_PLAY:

			Timer.Update(TDeltaTime);
			Replay::Instance().Update(TDeltaTime);

			ObjectManager::Instance().BeginFrame();

			Player->HandleInput();
			Physics::Instance().Update(TDeltaTime);
			ObjectManager::Instance().Update(TDeltaTime);
			Camera->Update(Player->GetNode()->getPosition());
			Interface::Instance().Update(TDeltaTime);
			Scripting::Instance().UpdateTimedCallbacks();

			ObjectManager::Instance().EndFrame();

			Replay::Instance().ResetNextPacketTimer();
		break;
		default:
		break;
	}
}

// Draws the current state
void PlayState::Draw() {
	
	switch(State) {
		case STATE_PLAY:
			stringw String = stringw(L"FPS: ") + stringw(irrDriver->getFPS()) + stringw("\nTime: ") + Interface::Instance().ConvertSecondsToString(Timer.GetTime());
			HUDText->setText(String.c_str());
		break;
	}

}

// Draws the HUD
void PlayState::InitPlay() {
	
	// Draw interface
	Interface::Instance().Clear();
	Level::Instance().SetClearColor();
	HUDText = irrGUI->addStaticText(L"FPS: 1", rect<s32>(10, 10, 160, 90), false, true, 0, -1, true);
	
	irrDevice->getCursorControl()->setVisible(false);
	Input::Instance().SetMouseLocked(true);

	State = STATE_PLAY;
}

// Draws the pause menu
void PlayState::InitPause() {
	
	// Draw interface
	Interface::Instance().Clear();
	IGUIButton *ButtonResume = irrGUI->addButton(Interface::Instance().GetAbsoluteRectWH(0.4f, 0.3f, 0.2f, 0.05f), 0, PAUSE_RESUME, L"Resume");
	IGUIButton *ButtonSaveReplay = irrGUI->addButton(Interface::Instance().GetAbsoluteRectWH(0.4f, 0.4f, 0.2f, 0.05f), 0, PAUSE_SAVEREPLAY, L"Save Replay");
	IGUIButton *ButtonMainMenu = irrGUI->addButton(Interface::Instance().GetAbsoluteRectWH(0.4f, 0.5f, 0.2f, 0.05f), 0, PAUSE_MAINMENU, L"Main Menu");
	
	irrDevice->getCursorControl()->setVisible(true);
	Input::Instance().SetMouseLocked(false);

	State = STATE_PAUSED;
}

// Draws the save replay GUI
void PlayState::InitSaveReplay() {

	// Draw interface
	Interface::Instance().Clear();
	IGUIEditBox *EditName = irrGUI->addEditBox(L"", Interface::Instance().GetAbsoluteRectWH(0.37f, 0.3f, 0.26f, 0.05f), true, 0, SAVEREPLAY_NAME);
	IGUIButton *ButtonSave = irrGUI->addButton(Interface::Instance().GetAbsoluteRectWH(0.37f, 0.4f, 0.1f, 0.05f), 0, SAVEREPLAY_SAVE, L"Save");
	IGUIButton *ButtonCancel = irrGUI->addButton(Interface::Instance().GetAbsoluteRectWH(0.53f, 0.4f, 0.1f, 0.05f), 0, SAVEREPLAY_CANCEL, L"Cancel");

	irrGUI->setFocus(EditName);
	EditName->setMax(32);

	State = STATE_SAVEREPLAY;
}

// Draws the lose screen
void PlayState::InitLose() {
	
	// Draw interface
	Interface::Instance().Clear();
	IGUIStaticText *TextWin = irrGUI->addStaticText(L"You Lose", Interface::Instance().GetCenteredTextRect(L"You Lose", 0.15f, 0.20f), false, false, 0, -1, false);
	IGUIButton *ButtonRestartLevel = irrGUI->addButton(Interface::Instance().GetAbsoluteRectWH(0.4f, 0.3f, 0.2f, 0.05f), 0, LOSE_RESTARTLEVEL, L"Restart Level");
	IGUIButton *ButtonSaveReplay = irrGUI->addButton(Interface::Instance().GetAbsoluteRectWH(0.4f, 0.4f, 0.2f, 0.05f), 0, LOSE_SAVEREPLAY, L"Save Replay");
	IGUIButton *ButtonMainMenu = irrGUI->addButton(Interface::Instance().GetAbsoluteRectWH(0.4f, 0.5f, 0.2f, 0.05f), 0, LOSE_MAINMENU, L"Main Menu");

	irrDevice->getCursorControl()->setVisible(true);
	Input::Instance().SetMouseLocked(false);

	NextLevel = CurrentLevel;

	State = STATE_LOSE;
}

// Draws the win screen
void PlayState::InitWin() {
	
	// Draw interface
	Interface::Instance().Clear();
	Graphics::Instance().SetClearColor(SColor(255, 0, 0, 0));
	Graphics::Instance().SetDrawScene(false);

	SaveLevelStruct Stats;
	Save::Instance().GetLevelStats(CurrentLevel, Stats);
	for(u32 i = 0; i < Stats.HighScores.size(); i++) {
		stringw DateString = stringw(i+1) + stringw(": ") + stringw(Stats.HighScores[i].Date.c_str());
		IGUIStaticText *Date = irrGUI->addStaticText(DateString.c_str(), Interface::Instance().GetAbsoluteRectWH(0.37f, 0.25f + i * 0.03f, 0.2f, 0.1f), false, false, 0, -1, false);
		IGUIStaticText *Time = irrGUI->addStaticText(Interface::Instance().ConvertSecondsToString(Stats.HighScores[i].Time).c_str(), Interface::Instance().GetAbsoluteRectWH(0.57f, 0.25f + i * 0.03f, 0.2f, 0.1f), false, false, 0, -1, false);
	}

	stringw TimeString = stringw(L"Time: ") + Interface::Instance().ConvertSecondsToString(Timer.GetTime());
	IGUIStaticText *TextWin = irrGUI->addStaticText(L"You Win", Interface::Instance().GetCenteredTextRect(L"You Win", 0.10f, 0.15f), false, false, 0, -1, true);
	IGUIStaticText *TextTime = irrGUI->addStaticText(TimeString.c_str(), Interface::Instance().GetCenteredTextRect(TimeString.c_str(), 0.15f, 0.20f), false, false, 0, -1, true);
	IGUIButton *ButtonRestartLevel = irrGUI->addButton(Interface::Instance().GetAbsoluteRectCentered(0.25f, 0.8f, 0.15f, 0.05f), 0, WIN_RESTARTLEVEL, L"Restart Level");
	IGUIButton *ButtonNextLevel = irrGUI->addButton(Interface::Instance().GetAbsoluteRectCentered(0.4f, 0.8f, 0.15f, 0.05f), 0, WIN_NEXTLEVEL, L"Next Level");
	IGUIButton *ButtonSaveReplay = irrGUI->addButton(Interface::Instance().GetAbsoluteRectCentered(0.55f, 0.8f, 0.15f, 0.05f), 0, WIN_SAVEREPLAY, L"Save Replay");
	IGUIButton *ButtonMainMenu = irrGUI->addButton(Interface::Instance().GetAbsoluteRectCentered(0.7f, 0.8f, 0.15f, 0.05f), 0, WIN_MAINMENU, L"Main Menu");
	
	if(NextLevel == "")
		ButtonNextLevel->setEnabled(false);

	irrDevice->getCursorControl()->setVisible(true);
	Input::Instance().SetMouseLocked(false);

	State = STATE_WIN;
}

// Saves a replay
void PlayState::SaveReplay() {

	IGUIEditBox *EditName = static_cast<IGUIEditBox *>(irrGUI->getRootGUIElement()->getElementFromId(SAVEREPLAY_NAME));
	if(EditName != NULL)
		Replay::Instance().SaveReplay(EditName->getText());

	switch(TargetState) {
		case STATE_WIN: {
			InitWin();
			
			IGUIButton *ButtonSaveReplay = static_cast<IGUIButton *>(irrGUI->getRootGUIElement()->getElementFromId(WIN_SAVEREPLAY));
			ButtonSaveReplay->setEnabled(false);
		}
		break;
		case STATE_LOSE: {
			InitLose();
			
			IGUIButton *ButtonSaveReplay = static_cast<IGUIButton *>(irrGUI->getRootGUIElement()->getElementFromId(LOSE_SAVEREPLAY));
			ButtonSaveReplay->setEnabled(false);
		}
		break;
		default:
			InitPause();
		break;
	}
}
