/*************************************************************************************
*	irrlamb - http://code.google.com/p/irrlamb/
*	Copyright (C) 2007  Alan Witkowski
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation, either version 3 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program.  If not, see <http://www.gnu.org/licenses/>.
**************************************************************************************/
#include "viewreplay.h"
#include "engine/globals.h"
#include "engine/input.h"
#include "engine/graphics.h"
#include "engine/config.h"
#include "engine/physics.h"
#include "engine/level.h"
#include "engine/objectmanager.h"
#include "engine/replay.h"
#include "engine/audio.h"
#include "engine/interface.h"
#include "engine/random.h"
#include "engine/camera.h"
#include "engine/game.h"
#include "objects/player.h"
#include "objects/orb.h"
#include "play.h"
#include "menu.h"

// Initializes the state
int ViewReplayState::Init() {

	Timer.Start();

	Graphics::Instance().SetDrawScene(false);
	ReplaySpeed = 1.0f;
	TimeAccumulator = 0.0f;

	// Add gui
	HUDText = irrGUI->addStaticText(L"FPS: 1", rect<s32>(10, 10, 200, 90), false, true, 0, -1, true);
	irrDevice->getCursorControl()->setVisible(false);
	Input::Instance().SetMouseLocked(true);

	// Set up physics world
	Physics::Instance().Init(false);

	// Load replay
	if(!Replay::Instance().StartReplay(CurrentReplay.c_str(), Header))
		return 0;
	Replay::Instance().ReadEvent(NextEvent);

	// Load the level
	if(!Level::Instance().Init(Header.LevelFile, false))
		return 0;

	// Get the player
	Player = static_cast<PlayerClass *>(ObjectManager::Instance().GetObjectByName("player0"));
	if(Player == NULL)
		return 0;

	// Add camera
	Camera = new CameraClass();

	return 1;
}

// Shuts the state down
int ViewReplayState::Close() {

	Replay::Instance().StopReplay();

	delete Camera;
	Level::Instance().Close();
	ObjectManager::Instance().ClearObjects();
	Interface::Instance().Clear();
	Physics::Instance().Close();
	irrScene->clear();
	
	return 1;
}

// Key presses
bool ViewReplayState::HandleKeyPress(EKEY_CODE TKey) {

	bool Processed = true;

	switch(TKey) {
		case KEY_ESCAPE:
			if(Game::Instance().GetQuitOnEscape())
				Game::Instance().SetDone(true);
			else
				Game::Instance().ChangeState(MenuState::Instance());
		break;
		case KEY_F1:
			Game::Instance().ChangeState(MenuState::Instance());
		break;
		case KEY_F12:
			Graphics::Instance().SaveScreenshot();					
		break;
		case KEY_RIGHT:
			Timer.Update(1000);
			Replay::Instance().Update(1000);
		break;
		case KEY_UP:
			ReplaySpeed += 0.25f;
			if(ReplaySpeed >= 10.0f)
				ReplaySpeed = 10.0f;
		break;
		case KEY_DOWN:
			ReplaySpeed -= 0.25f;
			if(ReplaySpeed <= 0.0f)
				ReplaySpeed = 0.0f;
		break;
		default:
			Processed = false;
		break;
	}

	return Processed;
}

// Mouse motion
void ViewReplayState::HandleMouseMotion(float TUpdateX, float TUpdateY) {

	Camera->HandleMouseMotion(TUpdateX, TUpdateY);
}

// Mouse buttons
bool ViewReplayState::HandleMousePress(int TButton, int TMouseX, int TMouseY) {

	return false;
}

// Mouse wheel
void ViewReplayState::HandleMouseWheel(float TDirection) {

}

// Updates the current state
void ViewReplayState::Update(u32 TDeltaTime) {

	// Update the replay
	TimeAccumulator += TDeltaTime * ReplaySpeed;
	if(TimeAccumulator >= 1.0f) {
		Timer.Update((u32)(TimeAccumulator));
		Replay::Instance().Update((u32)(TimeAccumulator));

		TimeAccumulator -= (u32)(TimeAccumulator);
	}

	while(!Replay::Instance().ReplayStopped() && Timer.GetTime() >= NextEvent.TimeStamp) {
		//printf("Processing header packet: time=%d\n", NextEvent.TimeStamp);
		
		switch(NextEvent.Type) {
			case ReplayClass::PACKET_MOVEMENT:
				ObjectManager::Instance().UpdateFromMovementPacket(NextEvent);
			break;
			case ReplayClass::PACKET_CREATE: {
				ReplayObjectCreateStruct Packet;
				Replay::Instance().ReadPacket(Packet);

				Level::Instance().SetDefaultPosition(Packet.Position);
				Level::Instance().SetDefaultRotation(Packet.Rotation);
				Level::Instance().CreateObjectFromTemplateIndex(Packet.Template);
			}
			break;
			case ReplayClass::PACKET_DELETE: {
				ReplayObjectDeleteStruct Packet;
				Replay::Instance().ReadPacket(Packet);
				ObjectManager::Instance().DeleteObjectByID(Packet.ObjectID);
			}
			break;
			case ReplayClass::PACKET_CAMERA: {
				ReplayCameraStruct Packet;
				Replay::Instance().ReadPacket(Packet);

				Camera->GetNode()->setPosition(Packet.Position);
				Camera->GetNode()->setTarget(Packet.LookAt);
				Graphics::Instance().SetDrawScene(true);
			}
			break;
			case ReplayClass::PACKET_ORBDEACTIVATE: {
				ReplayOrbDeactivateStruct Packet;
				Replay::Instance().ReadPacket(Packet);

				OrbClass *Orb = static_cast<OrbClass *>(ObjectManager::Instance().GetObjectByID(Packet.ObjectID));
				Orb->StartDeactivation("", Packet.Length);
			}
			break;
			default:
				irrLog->log("Bad replay event type\n");
			break;
		}

		Replay::Instance().ReadEvent(NextEvent);
	}
	
	ObjectManager::Instance().UpdateReplay(TDeltaTime);
	Interface::Instance().Update(TDeltaTime);
}

// Draws the current state
void ViewReplayState::Draw() {
	
	stringw String = stringw("FPS: ") + stringw(irrDriver->getFPS()) + stringw("\nTime: ");
	if(!Replay::Instance().ReplayStopped())
		String += Interface::Instance().ConvertSecondsToString(Timer.GetTime());
	else
		String += Interface::Instance().ConvertSecondsToString(Header.CompletionTime);

	char Buffer[10];
	sprintf(Buffer, "%.2f", ReplaySpeed);
	String += stringw("\nReplay Speed: ") + stringw(Buffer) + stringw("x");

	HUDText->setText(String.c_str());
}
